# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import Facility
from .. import Mixer
from ..._graphics import mixer_graphics

__all__ = ('BlowdownMixer',)

class BlowdownMixer(Facility, Mixer):
    network_priority = 2
    _graphics = mixer_graphics
    ticket_name = 'BDM'
    def __init__(self, ID='', ins=None, outs=(), thermo=None, rigorous=False):
        Facility.__init__(self, ID, ins, outs, thermo)
        self.rigorous = rigorous


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena and contributors <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

from .._unit import Unit

__all__ = ('Duplicator',)


class Duplicator(Unit, isabstract=True):
    """
    Create a Duplicator object that takes in one inlet stream
    and duplicates it to all outlet streams.

    Parameters
    ----------
    ins : stream
        Inlet stream.
    outs : stream sequence
        Duplicated outlet streams.

    Examples
    --------
    Create a Duplicator object with an ID and any number of outlet streams:

    >>> from biosteam import settings, Stream, units
    >>> settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
    >>> D1 = units.Duplicator('D1', ins=feed, outs=('out_a', 'out_b', 'out_c'))
    >>> D1.simulate()
    >>> D1.show()
    Duplicator: D1
    ins...
    [0] feed
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10
    outs...
    [0] out_a
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10
    [1] out_b
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10
    [2] out_c
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10

    """
    _N_outs = 2
    _outs_size_is_fixed = False
    
    def _load_stream_links(self):
        feed, = self.ins
        duplicated_outs = self.outs
        for stream in duplicated_outs:
            stream.link_with(feed)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from ._unit import Unit

__all__ = ('Facility',)

def get_network_priority(facility):
    return facility.network_priority

class Facility(Unit, isabstract=True,
               new_graphics=False):
    """Abstract class for facilities that are run after simulation of all
    unit operations within a system path."""
    autonumber = False # Default ID will not include number
    @staticmethod
    def ordered_facilities(facilities):
        """Return facilities ordered according to their network priority."""
        return sorted(facilities, key=get_network_priority)
    
    def __init_subclass__(cls,
                          isabstract=False,
                          new_graphics=True):
        super().__init_subclass__(isabstract,
                                  new_graphics)
        if not hasattr(cls, 'network_priority'):
            raise NotImplementedError('Facility subclasses must implement a '
                                      '`network_priority` attribute to designate '
                                      'the order of simulation relative to other '
                                      'facilities')
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        self._system = None
        self._other_units = None
    
    @property
    def other_units(self):
        return self._other_units
        



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .._unit import Unit
from thermosteam.separations import material_balance
from .._graphics import process_specification_graphics
from ..utils import static

__all__ = ('MassBalance',)

# %% Mass Balance Unit

@static
class MassBalance(Unit):
    """
    Create a Unit object that changes net input flow rates to satisfy output
    flow rates. This calculation is based on mass balance equations for
    specified IDs. 

    Parameters
    ----------
    ins : stream
        Inlet stream. Doesn't actually affect mass balance. It's just to
        show the position in the process.
    outs : stream
        Outlet stream. Doesn't actually affect mass balance. It's just to
        show the position in the process.
    chemical_IDs : tuple[str]
        Chemicals that will be used to solve mass balance linear equations.
        The number of chemicals must be same as the number of input streams varied.
    variable_inlets : Iterable[Stream]
        Inlet streams that can vary in net flow rate to accomodate for the
        mass balance.
    constant_inlets: Iterable[Stream], optional
        Inlet streams that cannot vary in flow rates.
    constant_outlets: Iterable[Stream], optional
        Outlet streams that cannot vary in flow rates.
    is_exact=True : bool, optional
        True if exact flow rate solution is required for the specified IDs.
    balance='flow' : {'flow', 'composition'}, optional
          * 'flow': Satisfy output flow rates
          * 'composition': Satisfy net output molar composition

    Examples
    --------
    MassBalance are Unit objects that serve to alter flow rates of selected
    chemicals and input streams to satisfy the mass balance.
    The example below uses the MassBalance object to satisfy the target
    flow rate feeding the mixer M1:
    
    >>> from biosteam import System, Stream, settings, main_flowsheet
    >>> from biosteam.units import (Mixer, Splitter, StorageTank, Pump,
    ...                             Flash, MassBalance)
    >>> main_flowsheet.set_flowsheet('mass_balance_example')
    >>> settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> water = Stream('water',
    ...                Water=40,
    ...                units='lb/s',
    ...                T=350, P=101325)
    >>> ethanol = Stream('ethanol',
    ...                  Ethanol=190, Water=30,
    ...                  T=300, P=101325)
    >>> target = Stream('target',
    ...                 Ethanol=500, Water=500)
    >>> T1 = StorageTank('T1', outs='s1')
    >>> T2 = StorageTank('T2', outs='s2')
    >>> P1 = Pump('P1', P=101325, outs='s3')
    >>> P2 = Pump('P2', P=101325, outs='s4')
    >>> M1 = Mixer('M1', outs='s5')
    >>> S1 = Splitter('S1', outs=('s6', 's7'), split=0.5)
    >>> F1 = Flash('F1', outs=('s8', 's9'), V=0.5, P =101325)
    >>> MB1 = MassBalance('MB1', outs='s6_2',
    ...                   variable_inlets=[water, ethanol],
    ...                   constant_inlets=[S1-0],
    ...                   constant_outlets=[target],
    ...                   chemical_IDs=('Ethanol', 'Water'),
    ...                   description='Adjust flow rate of feed to mixer')
    >>> # Connect units
    >>> water-T1-P1
    <Pump: P1>
    >>> ethanol-T2-P2
    <Pump: P2>
    >>> [P1-0, P2-0, MB1-0]-M1-F1-1-S1-0-MB1
    <MassBalance: MB1>
    >>> sys = main_flowsheet.create_system('sys')
    >>> # Make diagram to view system
    >>> # sys.diagram()
    >>> sys.simulate();
    >>> target.show()
    Stream: target
     phase: 'l', T: 298.15 K, P: 101325 Pa
     flow (kmol/hr): Water    500
                     Ethanol  500
    
    """
    _graphics = process_specification_graphics
    _N_heat_utilities = 0
    _N_ins = _N_outs = 1

    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 chemical_IDs=None, variable_inlets=(),
                 constant_outlets=(), constant_inlets=(),
                 is_exact=True, balance='flow',
                 description=""):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.variable_inlets = variable_inlets
        self.constant_inlets = constant_inlets
        self.constant_outlets = constant_outlets
        self.chemical_IDs = tuple(chemical_IDs)
        self.is_exact = is_exact
        self.balance = balance
        self.description = description
        
    def _run(self):
        material_balance(
            chemical_IDs=self.chemical_IDs,
            variable_inlets=self.variable_inlets,
            constant_outlets=self.constant_outlets, 
            constant_inlets=self.constant_inlets,
            is_exact=self.is_exact,
            balance=self.balance
        )


# %% Energy Balance Unit

# class EnergyBalance(Unit):
#     """Create a Unit object that changes a stream's temperature, flow rate, or vapor fraction to satisfy energy balance.

#     **Parameters**

#         **index:** [int] Index of stream that can vary in temperature, flow rate, or vapor fraction.
        
#         **Type:** [str] Should be one of the following
#             * 'T': Vary temperature of output stream
#             * 'F': Vary flow rate of input/output stream
#             * 'V': Vary vapor fraction of output stream
        
#         **Qin:** *[float]* Additional energy input.
        
#     .. Note:: This is not a mixer, input streams and output streams should match flow rates.

#     """
#     _kwargs = {'index': None,
#                'Type': 'T',
#                'Qin': 0}
#     line = 'Balance'
#     _has_cost = False
#     _graphics = MassBalance._graphics
#     _init_ins = MassBalance._init_ins
#     _init_outs = MassBalance._init_outs
    
#     def _run(self):        # Get arguments
#         ins = self.ins.copy()
#         outs = self.outs.copy()
#         kwargs = self._kwargs
#         index = kwargs['index']
#         Type = kwargs['Type']
#         Qin = kwargs['Qin']
        
#         # Pop out required streams
#         if Type == 'F':
#             s_in = ins.pop(index)
#             s_out = outs.pop(index)
#         else:
#             s = outs.pop(index)
        
#         # Find required enthalpy
#         H_in = sum(i.H for i in ins) + Qin
#         H_out = sum(o.H for o in outs)
#         H_s = H_out - H_in
        
#         # Set enthalpy
#         if Type == 'T':
#             s.H = -H_s
#         elif Type == 'V':
#             s.enable_phases()
#             s.VLE(Qin=s.H - H_s)
#         elif Type == 'F':
#             s.mol *= (s_out.H - s_in.H)/H_s
#         else:
#             raise ValueError(f"Type must be 'T', 'V' or 'F', not '{Type}'")
            
        
        



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .._unit import Unit
from .._graphics import junction_graphics
from .._power_utility import PowerUtility
from ..exceptions import UndefinedChemical
from ..utils.piping import Inlets, Outlets

__all__ = ('Junction',)


# %% Connect between different property packages

class Junction(Unit):
    """
    Create a Junction object that copies data from `upstream` to `downstream`. 
    This serves to connect streams with different property packages.
    
    Parameters
    ----------    
    upstream=None : Stream or str, defaults to missing stream
        Stream that will be copied to `downstream`.
    downstream="" : Stream or str, defaults to missing stream
        Flow rate, T, P, and phase information
        will be copied from `upstream` to this stream.
        If None, stream will be missing.
    
    Examples
    --------
    Create a Junction object and connect streams with different chemicals:
        
    >>> from biosteam import *
    >>> settings.set_thermo(['Water'])
    >>> s1 = Stream('s1', Water=20)
    >>> settings.set_thermo(['Ethanol', 'Water'], cache=True)
    >>> s2 = Stream('s2') # Note that s2 and s1 have different chemicals defined
    >>> J1 = units.Junction('J1', s1, s2)
    >>> J1.simulate()
    >>> J1.show()
    Junction: J1
    ins...
    [0] s1
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  20
    outs...
    [0] s2
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  20
        
    """
    _stacklevel = Unit._stacklevel
    _graphics = junction_graphics
    heat_utilities = ()
    power_utility = PowerUtility()
    baseline_purchase_cost = 0.
    baseline_purchase_costs = {}
    purchase_cost = 0.
    purchase_costs = {}
    installed_cost = 0.
    installed_costs = {}
    utility_cost = 0.
    
    def __init__(self, ID="", upstream=None, downstream=None, thermo=None):
        self._register(ID)
        thermo = self._load_thermo(thermo)
        self._init_specification()
        self._isdynamic = False
        self._ins = Inlets(self, 1, upstream, thermo, True, self._stacklevel)
        self._outs = Outlets(self, 1, downstream, thermo, True, self._stacklevel)
    
    @property
    def upstream(self):
        return self._ins[0]
    @upstream.setter
    def upstream(self, upstream):
        self._ins[0] = upstream
        
    @property
    def downstream(self):
        return self._outs[0]
    @downstream.setter
    def downstream(self, downstream):
        self._outs[0] = downstream
        
    def _setup(self): 
        pass
    
    def _run(self): 
        try: self._outs[0].copy_like(self._ins[0])
        except UndefinedChemical:
            self._reset_thermo(self._ins[0]._thermo)
            self._outs[0].copy_like(self._ins[0])
    simulate = Unit.run

    @property
    def inlet_utility_indices(self): return {}
    
    @property
    def outlet_utility_indices(self): return {}



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains unit operations for splitting flows.

.. contents:: :local:
    
.. autoclass:: biosteam.units.splitting.Splitter
.. autoclass:: biosteam.units.splitting.Splitter
.. autoclass:: biosteam.units.splitting.PhaseSplitter 
.. autoclass:: biosteam.units.splitting.MockSplitter
.. autoclass:: biosteam.units.splitting.ReversedSplitter

"""
from .. import Unit
from .._graphics import splitter_graphics
from thermosteam import separations

__all__ = ('Splitter', 'PhaseSplitter', 'FakeSplitter', 'MockSplitter',
           'ReversedSplitter')

class Splitter(Unit):
    """
    Create a splitter that separates mixed streams based on splits.

    Parameters
    ----------
    ins : stream
        Inlet fluid to be split.
    outs : stream sequence
        * [0] Split stream
        * [1] Remainder stream    
    split : Should be one of the following
        * [float] The fraction of net feed in the 0th outlet stream
        * [array_like] Componentwise split of feed to 0th outlet stream
        * [dict] ID-split pairs of feed to 0th outlet stream
    order=None : Iterable[str], defaults to biosteam.settings.chemicals.IDs
        Chemical order of split.
    
    Examples
    --------
    Create a Splitter object with an ID, a feed stream, two outlet streams,
    and an overall split:
        
    >>> from biosteam import units, settings, Stream
    >>> settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = Stream('feed', Water=20, Ethanol=10, T=340)
    >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'), split=0.1)
    >>> S1.simulate()
    >>> S1.show()
    Splitter: S1
    ins...
    [0] feed
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10
    outs...
    [0] top
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    2
                        Ethanol  1
    [1] bot
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    18
                        Ethanol  9
      
    Create a Splitter object, but this time with a componentwise split
    using a dictionary:
    
    >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
    ...                     split={'Water': 0.1, 'Ethanol': 0.99})
    >>> S1.simulate()
    >>> S1.show()
    Splitter: S1
    ins...
    [0] feed
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10
    outs...
    [0] top
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    2
                        Ethanol  9.9
    [1] bot
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    18
                        Ethanol  0.1
                           
    Create a Splitter object using componentwise split, but this time specify the order:
    
    >>> S1 = units.Splitter('S1', ins=feed, outs=('top', 'bot'),
    ...                     order=('Ethanol', 'Water'),
    ...                     split=(0.99, 0.10))
    >>> S1.simulate()
    >>> S1.show()
    Splitter: S1
    ins...
    [0] feed
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    20
                        Ethanol  10
    outs...
    [0] top
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    2
                        Ethanol  9.9
    [1] bot
        phase: 'l', T: 340 K, P: 101325 Pa
        flow (kmol/hr): Water    18
                        Ethanol  0.1

    Splits can also be altered after creating the splitter:
        
    >>> S1.split = 0.5
    >>> S1.isplit.show()
    SplitIndexer:
     Water    0.5
     Ethanol  0.5
     
    >>> S1.isplit['Water'] = 1.0
    >>> S1.isplit.show()
    SplitIndexer:
     Water    1
     Ethanol  0.5
     
    >>> S1.split = [0.9, 0.8]
    >>> S1.isplit.show()
    SplitIndexer:
     Water    0.9
     Ethanol  0.8

    """
    _N_outs = 2
    _graphics = splitter_graphics
    
    @property
    def isplit(self):
        """[ChemicalIndexer] Componentwise split of feed to 0th outlet stream."""
        return self._isplit
    @property
    def split(self):
        """[Array] Componentwise split of feed to 0th outlet stream."""
        return self._isplit._data
    @split.setter
    def split(self, values):
        split = self.split
        if split is not values:
            split[:] = values
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, split, order=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        self._isplit = self.thermo.chemicals.isplit(split, order)
        
    def _run(self):
        feed = self._ins[0]
        isplit = self._isplit
        if isplit.chemicals is not feed.chemicals: self._reset_thermo(feed._thermo)
        feed.split_to(*self.outs, isplit._data)


class PhaseSplitter(Unit):
    """
    Create a PhaseSplitter object that splits the feed to outlets by phase.
    
    Parameters
    ----------
    ins : stream
        Feed.
    outs : streams
        Outlets.
        
    Notes
    -----
    Phases allocate to outlets in alphabetical order. For example,
    if the feed.phases is 'gls' (i.e. gas, liquid, and solid), the phases
    of the outlets will be 'g', 'l', and 's'.
        
    Examples
    --------
    >>> import biosteam as bst
    >>> bst.settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = bst.Stream('feed', Water=10, Ethanol=10)
    >>> feed.vle(V=0.5, P=101325)
    >>> s1 = bst.Stream('s1')
    >>> s2 = bst.Stream('s2')
    >>> PS = bst.PhaseSplitter('PS', feed, [s1, s2])
    >>> PS.simulate()
    >>> PS.show()
    PhaseSplitter: PS
    ins...
    [0] feed
        phases: ('g', 'l'), T: 353.86 K, P: 101325 Pa
        flow (kmol/hr): (g) Water    3.86
                            Ethanol  6.14
                        (l) Water    6.14
                            Ethanol  3.86
    outs...
    [0] s1
        phase: 'g', T: 353.86 K, P: 101325 Pa
        flow (kmol/hr): Water    3.86
                        Ethanol  6.14
    [1] s2
        phase: 'l', T: 353.86 K, P: 101325 Pa
        flow (kmol/hr): Water    6.14
                        Ethanol  3.86
    
    """
    _N_ins = 1
    _N_outs = 2
    _graphics = splitter_graphics
    
    def _run(self):
        separations.phase_split(*self.ins, self.outs)


class MockSplitter(Unit):
    """
    Create a MockSplitter object that does nothing when simulated.
    """
    _graphics = Splitter._graphics
    _N_ins = 1
    _N_outs = 2
    _outs_size_is_fixed = False
    
    def _run(self): pass

MockSplitter.line = 'Splitter'
FakeSplitter = MockSplitter    


class ReversedSplitter(Unit):
    """
    Create a splitter that, when simulated, sets the inlet stream based 
    on outlet streams. Must have only one input stream. The outlet streams will
    have the same temperature, pressure and phase as the inlet.
    
    """
    _graphics = Splitter._graphics
    _N_ins = 1
    _N_outs = 2
    _outs_size_is_fixed = False
    power_utility = None
    heat_utilities = ()
    results = None
    
    def _run(self):
        inlet, = self.ins
        outlets = self.outs
        reversed_split(inlet, outlets)


def reversed_split(inlet, outlets):
    inlet.mol[:] = sum([i.mol for i in outlets])
    T = inlet.T
    P = inlet.P
    phase = inlet.phase
    for out in outlets:
        out.T = T
        out.P = P
        out.phase = phase 

  
    



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
.. contents:: :local:
    
.. autoclass:: biosteam.units.PelletMill
.. autoclass:: biosteam.units.BagassePelletMill

"""
from .decorators import cost
from .._unit import Unit
from warnings import warn

__all__ = ('PelletMill', 'BagassePelletMill')

# TODO: The pellet mill is carbon steel. Add material factors later
@cost('Flow rate', CE=567., lb=362.8739, ub=36287.39,
      S=1., units='kg/hr', n=0.11, cost=8659.2, BM=2.3)
class PelletMill(Unit):
    """
    Create a pellet mill for pelletizing solids.
    
    Parameters
    ----------
    ins : stream
    outs : stream 
    kW : float, optional
        Power requirement [kWh/kg]. Defaults to 0.01 based on a heuristic from 
        Perry's handbook.
    
    Notes
    -----
    Feed is pushed through holes in dies of various shapes. The friction of 
    the material in the die holes supplies the resistance necessary for 
    compaction. Adjustable knives shear the rodlike extrudates into pellets of 
    the desired length. Although several designs are in use, the most commonly 
    used pellet mills operate by applying power to the die and rotating it 
    around a freely turning roller with fixed horizontal or vertical axis. 
    
    """
    _N_ins = 1
    _N_outs = 1
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, kW=0.01):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.kW = kW
        
    def _cost(self):
        self._decorated_cost()
        self.power_utility.consumption = self.kW * self.design_results['Flow rate']
        

class BagassePelletMill(PelletMill):
    """
    Create a pellet mill for pelletizing dryed bagasse.
    
    Parameters
    ----------
    ins : stream
    outs : stream 
    kW : float, optional
        User enforced power requirement [kWh/kg]. Defaults according to
        the `capacity_to_power` class attribute.
    
    Notes
    -----
    Feed is pushed through holes in dies of various shapes. The friction of 
    the material in the die holes supplies the resistance necessary for 
    compaction. Adjustable knives shear the rodlike extrudates into pellets of 
    the desired length. Although several designs are in use, the most commonly 
    used pellet mills operate by applying power to the die and rotating it 
    around a freely turning roller with fixed horizontal or vertical axis. 
    
    """
    
    #: Tuple[Tuple[float, float], float] Capacity ranges [kg/hr] and default 
    #: power requirements [kW]. Values from industrial supplier:
    #: http://www.biopelletmill.com/large-biomass-pellet-mill.html
    #: http://www.biopelletmill.com/small-biomass-pellet-mill.html
    capacity_to_power = [
        ((50, 100), 5.5),
        ((80, 120), 7.5),
        ((120, 200), 11),
        ((160, 250), 15),
        ((250, 400), 22),
        ((400, 800), 61.5),
        ((800, 1500), 94.5),
        ((1500, 2000), 114.5),
    ]
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, kW=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.kW = kW
        
    def default_kW(self, F_mass):
        for (lb, ub), kW in self.capacity_to_power:
            if F_mass < ub: return kW
        return F_mass * kW / ub # Extrapolate
        
    def _cost(self):
        self._decorated_cost()
        F_mass = self.design_results['Flow rate']
        self.power_utility.consumption = self.default_kW(F_mass)
        
        

    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>,
#                          Yalin Li <mailto.yalin.li@gmail.com>
#
# This module implements a filtering feature from the stats module of the QSDsan library:
# QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
# Copyright (C) 2020-, Yalin Li <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
from scipy.spatial.distance import cdist
import numpy as np
import pandas as pd
from ._state import State
from ._metric import Metric
from ._parameter import Parameter
from ._utils import var_indices, var_columns, indices_to_multiindex
from biosteam.exceptions import FailedEvaluation
from warnings import warn
from collections.abc import Sized
from biosteam.utils import TicToc
import pickle

__all__ = ('Model',)


# %% Grid of simulation blocks

class Model(State):
    """
    Create a Model object that allows for evaluation over a sample space.
    
    Parameters
    ----------
    system : System
        Should reflect the model state.
    metrics : tuple[Metric]
        Metrics to be evaluated by model.
    specification=None : Function, optional
        Loads specifications once all parameters are set. Specification should 
        simulate the system as well.
    params=None : Iterable[Parameter], optional
        Parameters to sample from.
    exception_hook : callable(exception, sample)
        Function called after a failed evaluation. The exception hook should 
        return either None or metric values given the exception and sample.

    """
    __slots__ = (
        'table',            # [DataFrame] All arguments and results.
        'retry_evaluation', # [bool] Whether to retry evaluation if it fails
        '_metrics',         # tuple[Metric] Metrics to be evaluated by model.
        '_index',           # list[int] Order of sample evaluation for performance.
        '_samples',         # [array] Argument sample space.
        '_exception_hook',  # [callable(exception, sample)] Should return either None or metric value given an exception and the sample.
    )
    def __init__(self, system, metrics=None, specification=None, 
                 parameters=None, retry_evaluation=True, exception_hook='warn'):
        super().__init__(system, specification, parameters)
        self.metrics = metrics or ()
        self.exception_hook = exception_hook
        self.retry_evaluation = retry_evaluation
        self.table = None
        self._erase()
        
    def copy(self):
        """Return copy."""
        copy = super().copy()
        copy._metrics = self._metrics
        if self.table is None:
            copy._samples = copy.table = None
        else:
            copy.table = self.table.copy()
            copy._samples = self._samples
            copy._index = self._index
        return copy
    
    def _erase(self):
        """Erase cached data."""
        self._samples = None
    
    @property
    def exception_hook(self):
        """
        [callable(exception, sample)] Function called after a failed 
        evaluation. The exception hook should return either None or metric 
        values given the exception and sample.
        
        """
        return self._exception_hook
    @exception_hook.setter
    def exception_hook(self, exception_hook):
        if not exception_hook or callable(exception_hook):
            self._exception_hook = exception_hook
            return
        if isinstance(exception_hook, str):
            if exception_hook == 'ignore':
                self._exception_hook = lambda exception, sample: None
            elif exception_hook == 'warn':
                self._exception_hook = lambda exception, sample: warn(FailedEvaluation(f"[{type(exception).__name__}] {exception}"), stacklevel=6)
            elif exception_hook == 'raise':
                def raise_exception(exception, sample): raise exception
                self._exception_hook = raise_exception
            else:
                raise ValueError(f"invalid exception hook name '{exception_hook}'; "
                                  "valid names include 'ignore', 'warn', and 'raise'")
        else:
            raise ValueError('exception hook must be either a callable, a string, or None')
    
    @property
    def metrics(self):
        """tuple[Metric] Metrics to be evaluated by model."""
        return tuple(self._metrics)
    @metrics.setter
    def metrics(self, metrics):
        metrics = list(metrics)
        isa = isinstance
        for i in metrics:
            if not isa(i, Metric):
                raise ValueError(f"metrics must be '{Metric.__name__}' "
                                 f"objects, not '{type(i).__name__}'")
        Metric.check_indices_unique(metrics)
        self._metrics = metrics
    
    def metric(self, getter=None, name=None, units=None, element='Biorefinery'):
        """
        Define and register metric.
        
        Parameters
        ----------    
        getter : function, optional
                 Should return metric.
        name : str, optional
               Name of metric. If None, defaults to the name of the getter.
        units : str, optional
                Metric units of measure
        element : object, optional
                  Element being evaluated. Works mainly for bookkeeping. 
                  Defaults to 'Biorefinery'.
        
        Notes
        -----
        This method works as a decorator.
        
        """
        if not getter: return lambda getter: self.metric(getter, name, units, element)
        metric = Metric(name, getter, units, element)
        Metric.check_index_unique(metric, self._metrics)
        self._metrics.append(metric)
        return metric 
    
    def _sample_hook(self, samples, parameters):
        if any([p.hook for p in parameters]):
            return np.array(
                [[(i if p.hook is None else p.hook(i))
                  for p, i in zip(parameters, row)]
                 for row in samples]
            )
        else:
            return samples
    
    def _load_sample_order(self, samples, parameters, ss, distance):
        """
        Sort simulation order to optimize convergence speed
        by minimizing perturbations to the system between simulations.
        This function may take long depending on the number of parameters 
        because it uses single-point sensitivity to inform the sorting 
        algorithm.
        
        """
        if distance is None: distance = 'cityblock'
        length = samples.shape[0]
        original_parameters = self._parameters
        columns = [i for i, parameter in enumerate(self._parameters) if parameter.kind == 'coupled']
        parameters = [parameters[i] for i in columns]
        samples = samples.copy()
        samples = samples[:, columns]
        samples_min = samples.min(axis=0)
        samples_max = samples.max(axis=0)
        samples_diff = samples_max - samples_min
        normalized_samples = (samples - samples_min) / samples_diff
        if ss: 
            def evaluate(sample, material_data):
                try:
                    self._parameters = parameters
                    for f, s in zip(self._parameters, sample): 
                        f.setter(s if f.scale is None else f.scale * s)
                    diff = self._system.converge(material_data=material_data)
                finally:
                    self._parameters = original_parameters
                return diff
            sample = [i.baseline for i in parameters]
            N_parameters = len(parameters)
            index = range(N_parameters)
            evaluate(sample)
            material_data = self._system.get_material_data()
            N_elements = material_data.material_flows.size
            diffs = np.zeros([N_parameters, N_elements])
            for i in index:
                sample_lb = sample.copy()
                sample_ub = sample.copy()
                lb = samples_min[i]
                ub = samples_max[i]
                hook = parameters[i].hook
                if hook:
                    sample_lb[i] = hook(lb)
                    sample_ub[i] = hook(ub)
                else:
                    sample_lb[i] = lb
                    sample_ub[i] = ub
                lb = evaluate(sample_lb, material_data).flatten()
                ub = evaluate(sample_ub, material_data).flatten()
                diffs[i] = ub - lb
            div = diffs.max(axis=0)
            div[div == 0.] = 1
            diffs /= div
            normalized_samples = normalized_samples @ diffs
        nearest_arr = cdist(normalized_samples, normalized_samples, metric=distance)
        nearest_arr = np.argsort(nearest_arr, axis=1)
        remaining = set(range(length))
        self._index = index = [0]
        nearest = nearest_arr[0, 1]
        index.append(nearest)
        remaining.remove(0)
        remaining.remove(nearest)
        N_remaining = length - 2
        N_remaining_last = N_remaining + 1
        while N_remaining:
            assert N_remaining_last != N_remaining, "issue in sorting algorithm"
            N_remaining_last = N_remaining
            for i in range(1, length):
                next_nearest = nearest_arr[nearest, i]
                if next_nearest in remaining:
                    nearest = next_nearest
                    remaining.remove(nearest)
                    index.append(nearest)
                    N_remaining -= 1
                    break
        
    def load_samples(self, samples=None, optimize=None, ss=None, 
                     file=None, autoload=None, autosave=None, distance=None):
        """
        Load samples for evaluation.
        
        Parameters
        ----------
        samples : numpy.ndarray, dim=2, optional
            All parameter samples to evaluate.
        optimize : bool, optional
            Whether to internally sort the samples to optimize convergence speed
            by minimizing perturbations to the system between simulations.
            Defaults to False.
        ss : bool, optional
            Whether to use single point sensitivity to inform the sorting algorithm. 
            Defaults to False.
        file : str, optional
            File to load/save samples and simulation order to/from.
        autosave : bool, optional
            Whether to save samples and simulation order to file (when not loaded from file).
        autoload : bool, optional
            Whether to load samples and simulation order from file (if possible).
        distance : str, optional
            Distance metric used for sorting. Defaults to 'cityblock'.
            See scipy.spatial.distance.cdist for options.
        
        Warning
        -------
        Depending on the number of parameters, optimizing sample simulation order 
        using single-point sensitivity may take long.
        
        """
        parameters = self._parameters
        Parameter.check_indices_unique(parameters)
        if autoload:
            try:
                with open(file, "rb") as f: (self._samples, self._index) = pickle.load(f)
            except FileNotFoundError: pass
            else:
                if (samples is None or (samples.shape == self._samples.shape and (samples == self._samples).all())):
                    metrics = self._metrics
                    empty_metric_data = np.zeros((len(samples), len(metrics)))
                    self.table = pd.DataFrame(np.hstack((samples, empty_metric_data)),
                                              columns=var_columns(parameters + metrics),
                                              dtype=float)
                    return 
        if not isinstance(samples, np.ndarray):
            raise TypeError(f'samples must be an ndarray, not a {type(samples).__name__} object')
        if samples.ndim == 1:
            samples = samples[:, np.newaxis]
        elif samples.ndim != 2:
            raise ValueError('samples must be 2 dimensional')
        N_parameters = len(parameters)
        if samples.shape[1] != N_parameters:
            raise ValueError(f'number of parameters in samples ({samples.shape[1]}) must be equal to the number of parameters ({N_parameters})')
        metrics = self._metrics
        samples = self._sample_hook(samples, parameters)
        if optimize: 
            self._load_sample_order(samples, parameters, ss, distance)
        else:
            self._index = list(range(samples.shape[0]))
        empty_metric_data = np.zeros((len(samples), len(metrics)))
        self.table = pd.DataFrame(np.hstack((samples, empty_metric_data)),
                                  columns=var_columns(parameters + metrics),
                                  dtype=float)
        self._samples = samples
        if autosave:
            obj = (samples, self._index)
            try:
                with open(file, 'wb') as f: pickle.dump(obj, f)
            except FileNotFoundError:
                import os
                head, tail = os.path.split(file)
                os.mkdir(head)
                with open(file, 'wb') as f: pickle.dump(obj, f)
            
    def single_point_sensitivity(self, 
            etol=0.01, array=False, parameters=None, metrics=None, evaluate=None, 
            **kwargs
        ):
        if parameters is None: parameters = self.parameters
        bounds = [i.bounds for i in parameters]
        sample = [i.baseline for i in parameters]
        N_parameters = len(parameters)
        index = range(N_parameters)
        if metrics is None: metrics = self.metrics
        N_metrics = len(metrics)
        values_lb = np.zeros([N_parameters, N_metrics])
        values_ub = values_lb.copy()
        if evaluate is None: evaluate = self._evaluate_sample
        baseline_1 = np.array(evaluate(sample, **kwargs))
        sys = self.system
        if not sys.isdynamic: kwargs['material_data'] = sys.get_material_data()
        for i in index:
            sample_lb = sample.copy()
            sample_ub = sample.copy()
            lb, ub = bounds[i]
            hook = parameters[i].hook
            if hook:
                sample_lb[i] = hook(lb)
                sample_ub[i] = hook(ub)
            else:
                sample_lb[i] = lb
                sample_ub[i] = ub
            values_lb[i, :] = evaluate(sample_lb, **kwargs)
            values_ub[i, :] = evaluate(sample_ub, **kwargs)
        baseline_2 = np.array(evaluate(sample, **kwargs))
        error = np.abs(baseline_2 - baseline_1)
        index, = np.where(error > 1e-6)
        error = error[index]
        relative_error = error / np.maximum.reduce([np.abs(baseline_1[index]), np.abs(baseline_2[index])])
        for i, idx in enumerate(index):
            if relative_error[i] > etol:
                raise RuntimeError(
                    f"inconsistent model; {metrics[idx]} has a value of "
                    f"{baseline_1[idx]} before evaluating sensitivity and "
                    f"{baseline_2[idx]} after"
                )
        baseline = 0.5 * (baseline_1 + baseline_2)
        if array:
            return baseline, values_lb, values_ub
        else:
            metric_index = var_columns(metrics)
            baseline = pd.Series(baseline, index=metric_index)
            df_lb = pd.DataFrame(values_lb, 
                                index=var_columns(parameters),
                                columns=metric_index)
            df_ub = df_lb.copy()
            df_ub[:] = values_ub
            return baseline, df_lb, df_ub
    
    def evaluate(self, notify=0, file=None, autosave=0, autoload=False,
                 **kwargs):
        """
        Evaluate metrics over the loaded samples and save values to `table`.
        
        Parameters
        ----------
        notify=0 : int, optional
            If 1 or greater, notify elapsed time after the given number of sample evaluations. 
        file : str, optional
            Name of file to save/load pickled evaluation results.
        autosave : int, optional
            If 1 or greater, save pickled evaluation results after the given number of sample evaluations.
        autoload : bool, optional
            Whether to load pickled evaluation results from file.
        kwargs : dict
            Any keyword arguments passed to :func:`biosteam.System.simulate`.
        
        Warning
        -------
        Any changes made to either the model or the samples will not be accounted
        for when autoloading and may lead to misleading results.
        
        """
        samples = self._samples
        if samples is None: raise RuntimeError('must load samples before evaluating')
        evaluate_sample = self._evaluate_sample
        table = self.table
        if notify:
            timer = TicToc()
            timer.tic()
            count = [0]
            def evaluate(sample, count=count, **kwargs):
                count[0] += 1
                values = evaluate_sample(sample, **kwargs)
                if not count[0] % notify:
                    print(f"{count} Elapsed time: {timer.elapsed_time:.0f} sec")
                return values
        else:
            evaluate = evaluate_sample
        if autoload: 
            try:
                with open(file, "rb") as f:
                    number, values, table_index, table_columns = pickle.load(f)
                if (table_index != table.index).any() or (table_columns != table.columns).any():
                    raise ValueError('table layout does not match autoload file')
                del table_index, table_columns
                index = self._index[number:]
            except:
                number = 0
                index = self._index
                values = [None] * len(index)   
            else:
                if notify: count[0] = number
        else:
            number = 0
            index = self._index
            values = [None] * len(index)
        
        export = 'export_state_to' in kwargs
        layout = table.index, table.columns
        for number, i in enumerate(index, number + 1): 
            if export: kwargs['sample_id'] = i
            values[i] = evaluate(samples[i], **kwargs)
            if autosave and not number % autosave: 
                obj = (number, values, *layout)
                try:
                    with open(file, 'wb') as f: pickle.dump(obj, f)
                except FileNotFoundError:
                    import os
                    head, tail = os.path.split(file)
                    os.mkdir(head)
                    with open(file, 'wb') as f: pickle.dump(obj, f)
        table[var_indices(self._metrics)] = values
    
    def _evaluate_sample(self, sample, **kwargs):
        state_updated = False
        try:
            self._update_state(sample, **kwargs)
            state_updated = True
            return [i() for i in self.metrics]
        except Exception as exception:
            self._reset_system()
            if self.retry_evaluation and state_updated:
                try:
                    self._update_state(sample, **kwargs)
                    self._specification() if self._specification else self._system.simulate(**kwargs)
                    return [i() for i in self.metrics]
                except Exception as new_exception: 
                    if self._exception_hook: 
                        values = self._exception_hook(new_exception, sample)
                        self._reset_system()
                        if isinstance(values, Sized) and len(values) == len(self.metrics):
                            return values
                        elif values is not None:
                            raise RuntimeError('exception hook must return either None or '
                                               'an array of metric values for the given sample')
            if self._exception_hook: 
                values = self._exception_hook(exception, sample)
                self._reset_system()
                if isinstance(values, Sized) and len(values) == len(self.metrics):
                    return values
                elif values is not None:
                    raise RuntimeError('exception hook must return either None or '
                                       'an array of metric values for the given sample')
            return self._failed_evaluation()
    
    def _reset_system(self):
        self._system.empty_recycles()
        self._system.reset_cache()
    
    def _failed_evaluation(self):
        self._reset_system()
        return [np.nan] * len(self.metrics)
    
    def metrics_at_baseline(self):
        """Return metric values at baseline sample."""
        baseline = self.get_baseline_sample()
        return self(baseline)
    
    def evaluate_across_coordinate(self, name, f_coordinate, coordinate,
                                   *, xlfile=None, notify=0, notify_coordinate=True,
                                   multi_coordinate=False, 
                                   f_evaluate=None):
        """
        Evaluate across coordinate and save sample metrics.
        
        Parameters
        ----------
        name : str or tuple[str]
            Name of coordinate
        f_coordinate : function
            Should change state of system given the coordinate.
        coordinate : array
            Coordinate values.
        xlfile : str, optional
            Name of file to save. File must end with ".xlsx"
        rule : str, optional
            Sampling rule. Defaults to 'L'.
        notify : bool, optional
            If True, notify elapsed time after each coordinate evaluation. Defaults to True.
        f_evaluate : callable, optional
            Function to evaluate model. Defaults to evaluate method.
        
        """
        if self._samples is None: raise RuntimeError('must load samples before evaluating')
        table = self.table
        N_samples, N_parameters = table.shape
        N_points = len(coordinate)
        
        # Initialize data containers
        metric_indices = var_indices(self.metrics)
        shape = (N_samples, N_points)
        metric_data = {i: np.zeros(shape) for i in metric_indices}
        if f_evaluate is None: f_evaluate = self.evaluate
        
        # Initialize timer
        if notify_coordinate:
            from biosteam.utils import TicToc
            timer = TicToc()
            timer.tic()
            def evaluate():
                f_evaluate(notify=notify)
                print(f"[Coordinate {n}] Elapsed time: {timer.elapsed_time:.0f} sec")
        else:
            evaluate = f_evaluate
        
        for n, x in enumerate(coordinate):
            f_coordinate(*x) if multi_coordinate else f_coordinate(x)
            evaluate()
            for metric in metric_data:
                metric_data[metric][:, n] = table[metric]
        
        if xlfile:
            if multi_coordinate:
                columns = pd.MultiIndex.from_tuples(coordinate,
                                                    names=name)
            else:
                columns = pd.Index(coordinate, name=name)
            
            # Save data to excel
            data = pd.DataFrame(data=np.zeros([N_samples, N_points]),
                                columns=columns)
            
            with pd.ExcelWriter(xlfile) as writer:
                for metric in self.metrics:
                    data[:] = metric_data[metric.index]
                    data.to_excel(writer, sheet_name=metric.short_description)
        return metric_data
    
    def spearman(self, parameters=None, metrics=None):
        warn(DeprecationWarning('this method will be deprecated in biosteam 2.25; '
                                'use spearman_r instead'), stacklevel=2)
        return self.spearman_r(parameters, metrics)[0]
    
    def spearman_r(self, parameters=None, metrics=None, filter=None, **kwargs): # pragma: no cover
        """
        Return two DataFrame objects of Spearman's rho and p-values between metrics 
        and parameters.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with metrics. Defaults to all parameters.
        metrics : Iterable[Metric], optional 
            Metrics to be correlated with parameters. Defaults to all metrics.
        
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.spearmanr`.
        
        See Also
        --------
        :func:`scipy.stats.spearmanr`
        
        """
        from scipy.stats import spearmanr
        return self._correlation(spearmanr, parameters, metrics, filter, kwargs)
    
    def pearson_r(self, parameters=None, metrics=None, filter=None, **kwargs):
        """
        Return two DataFrame objects of Pearson's rho and p-values between metrics 
        and parameters.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with metrics. Defaults to all parameters.
        metrics : Iterable[Metric], optional 
            Metrics to be correlated with parameters. Defaults to all metrics.
            
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.pearsonr`.
        
        See Also
        --------
        :func:`scipy.stats.pearsonr`
        
        """
        from scipy.stats import pearsonr
        return self._correlation(pearsonr, parameters, metrics, filter, kwargs)
    
    def kendall_tau(self, parameters=None, metrics=None, filter=None, **kwargs):
        """
        Return two DataFrame objects of Kendall's tau and p-values between metrics 
        and parameters.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with metrics. Defaults to all parameters.
        metrics : Iterable[Metric], optional 
            Metrics to be correlated with parameters. Defaults to all metrics.
            
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
            
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
        
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.kendalltau`.
        
        See Also
        --------
        :func:`scipy.stats.kendalltau`
        
        """
        from scipy.stats import kendalltau
        return self._correlation(kendalltau, parameters, metrics, filter, kwargs)
    
    def kolmogorov_smirnov_d(self, parameters=None, metrics=None, thresholds=[],
                             filter=None, **kwargs):
        """
        Return two DataFrame objects of Kolmogorov–Smirnov's D and p-values
        with the given thresholds for the metrics.
        
        For one particular parameter, all of the sampled values will be divided into two sets,
        one where the resulting metric value is smaller than or equal to the threshold,
        and the other where the resulting value is larger than the threshold.
        
        Kolmogorov–Smirnov test will then be performed for these two sets of values
        for the particular parameter.
        
        Parameters
        ----------
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with metrics. Defaults to all parameters.
        metrics : Iterable[Metric], optional 
            Metrics to be correlated with parameters. Defaults to all metrics.
        thresholds : Iterable[float]
            The thresholds for separating parameters into sets.
            
        Other Parameters
        ----------------
        filter : Callable(x, y) -> x, y, or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        **kwargs :
            Keyword arguments passed to :func:`scipy.stats.kstest`.
        
        See Also
        --------
        :func:`scipy.stats.kstest`
        
        """
        from scipy.stats import kstest
        metrics = metrics or self.metrics
        if len(thresholds) != len(metrics):
            raise ValueError(f'The number of metrics {len(metrics)} must match '
                             f'the number of thresholds ({len(thresholds)}).')
        kwargs['thresholds'] = thresholds
        return self._correlation(kstest, parameters, metrics, filter, kwargs)
    
    def _correlation(self, f, parameters, metrics, filter, kwargs):
        """
        Return two DataFrame objects of statistics and p-values between metrics 
        and parameters.
        
        Parameters
        ----------
        f : Callable
            Function with signature f(x, y) -> stat, p
        parameters : Iterable[Parameter], optional
            Parameters to be correlated with metrics. Defaults to all parameters.
        metrics : Iterable[Metric], optional 
            Metrics to be correlated with parameters. Defaults to all metrics.
        filter : Callable or string, optional
            Function that accepts 1d arrays of x and y values and returns 
            filtered x and y values to correlate. May also
            be one of the following strings:
                
            * 'none': no filter
            
            * 'omit nan': all NaN values are ignored in correlation
            
            * 'propagate nan': NaN values return NaN correlation results
            
            * 'raise nan': NaN values will raise a ValueError
            
        kwargs : dict
            Keyword arguments passed to `f`.
            
        """
        if not parameters: parameters = self._parameters
        table = self.table
        values = table.values.transpose()
        index = table.columns.get_loc
        parameter_indices = var_indices(parameters)
        parameter_data = [values[index(i)] for i in parameter_indices]
        metric_indices = var_indices(metrics or self.metrics)
        metric_data = [values[index(i)] for i in metric_indices]
                
        if not filter: filter = 'propagate nan'
        if isinstance(filter, str):
            name = filter.lower()
            if name == 'omit nan':
                def corr(x, y):
                    index = ~(np.isnan(x) | np.isnan(y))
                    return f(x[index], y[index], **kwargs)
            elif name == 'propagate nan':
                def corr(x, y):
                    if np.isnan(x).any() or np.isnan(y).any(): 
                        NaN = float('NaN')
                        return NaN, NaN
                    else:
                        return f(x, y, **kwargs)
            elif name == 'raise nan':
                def corr(x, y):
                    if np.isnan(x).any() or np.isnan(y).any():
                        raise ValueError('table entries contain NaN values')
                    return f(x, y)
            elif name == 'none':
                corr = lambda x, y: f(x, y, **kwargs)
            else: #: pragma: no cover
                raise ValueError(
                    f"invalid filter '{filter}'; valid filter names are: "
                    "'omit nan', 'propagate nan', 'raise nan', and 'none'"
                )
        elif callable(filter): 
            corr = lambda x, y: f(*filter(x, y), **kwargs)
        else: #: pragma: no cover
            raise TypeError("filter must be either a string or a callable; "
                            "not a '{type(filter).__name__}' object")

        if 'thresholds' not in kwargs:
            data = np.array([[corr(p, m) for m in metric_data] for p in parameter_data])
        else: # KS test
            thresholds = kwargs.pop('thresholds')
            data = np.array(
                [
                    [corr(parameter_data[n_p][metric_data[n_m]<=thresholds[n_m]],
                          parameter_data[n_p][metric_data[n_m]>thresholds[n_m]]) \
                     for n_m in range(len(metric_data))] \
                for n_p in range(len(parameter_data))])
        
        index = indices_to_multiindex(parameter_indices, ('Element', 'Parameter'))
        columns = indices_to_multiindex(metric_indices, ('Element', 'Metric'))        
        return [pd.DataFrame(i, index=index, columns=columns) for i in (data[..., 0], data[..., 1])]
        
    def create_fitted_model(self, parameters, metrics): # pragma: no cover
        from pipeml import FittedModel
        Xdf = self.table[[i.index for i in parameters]]
        ydf_index = metrics.index if isinstance(metrics, Metric) else [i.index for i in metrics]
        ydf = self.table[ydf_index]
        return FittedModel.from_dfs(Xdf, ydf)
    
    def __call__(self, sample):
        """Return pandas Series of metric values at given sample."""
        super().__call__(sample)
        return pd.Series({i.index: i() for i in self._metrics})
    
    def _repr(self):
        clsname = type(self).__name__
        newline = "\n" + " "*(len(clsname)+2)
        return f'{clsname}: {newline.join([i.describe() for i in self.metrics])}'
        
    def __repr__(self):
        return f'<{type(self).__name__}: {", ".join([i.name for i in self.metrics])}>'
    
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import biosteam as bst
from numpy.testing import assert_allclose


def test_isenthalpic_methane_liquefaction():
    thermo = bst.Thermo([bst.Chemical('CH4')])
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo, cache=True)
    feed = bst.Stream(CH4=1, T=-63 + 273.15, P=100e5, phase='g')
    V = bst.units.IsenthalpicValve(ins=feed, P=1e5, vle=True)
    V.simulate()
    # check outlet state
    out = V.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P, out.H],
        [0.8353, 0.1646, 111.474, 1e5, feed.H],
        rtol=1e-2,
    )
    pass


if __name__ == '__main__':
    test_isenthalpic_methane_liquefaction()



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains unit operations for drying solids.

.. contents:: :local:
    
.. autoclass:: biosteam.units.drying.DrumDryer


"""
import biosteam as bst
import flexsolve as flx
import numpy as np
from thermosteam import separations as sep
from .design_tools import (
    CEPCI_by_year,
    cylinder_diameter_from_volume, 
    cylinder_area,
)
from .decorators import cost
from .._unit import Unit

__all__ = ('DrumDryer', 'ThermalOxidizer')

# TODO: The drum dryer is carbon steel. Add material factors later
@cost('Peripheral drum area', CE=CEPCI_by_year[2007], ub=7854.0, BM=2.06,
      S=1235.35, units='m2', n=0.6, cost=0.52 * 2268000., kW=938.866)
class DrumDryer(Unit):
    """
    Create a drum dryer that dries solids by passing hot air 
    (heated by burning natural gas).
    
    Parameters
    ----------
    ins : stream sequence
        * [0] Wet solids.
        * [1] Air.
        * [2] Natural gas.
    outs : stream sequence
        * [0] Dried solids
        * [1] Hot air
        * [2] Emissions
    split : dict[str, float]
        Component splits to hot air (stream [1]).
    R : float, optional
        Flow of hot air over evaporation. Defaults to 1.4 wt gas / wt evap.
    H : float, optional
        Specific evaporation rate [kg/hr/m3]. Defaults to 20. 
    length_to_diameter : float, optional
        Note that the drum is horizontal. Defaults to 25.
    T : float, optional
        Operating temperature [K]. Defaults to 343.15.
    moisture_content : float
        Moisture content of solids [wt / wt]. Defaults to 0.10.
        
    Notes
    -----
    The flow rate for air in the inlet is varied to meet the `R` specification
    (i.e. flow of hot air over flow rate evaporated). The flow rate of inlet natural
    gas is also altered to meet the heat demand.
    
    The default parameter values are based on heuristics for drying 
    dried distillers grains with solubles (DDGS).
    
    """
    # auxiliary_unit_names = ('heat_exchanger',)
    _units = {'Evaporation': 'kg/hr',
              'Peripheral drum area': 'm2',
              'Diameter': 'm'}
    _N_ins = 3
    _N_outs = 3
    
    @property
    def isplit(self):
        """[ChemicalIndexer] Componentwise split of feed to 0th outlet stream."""
        return self._isplit
    @property
    def split(self):
        """[Array] Componentwise split of feed to 0th outlet stream."""
        return self._isplit._data
    
    @property
    def natural_gas(self):
        """[Stream] Natural gas to satisfy steam and electricity requirements."""
        return self.ins[2]
    
    def __init__(self, ID="", ins=None, outs=(), thermo=None, *,
                 split, R=1.4, H=20., length_to_diameter=25, T=343.15, P=10*101325,
                 moisture_content=0.15, utility_agent='Natural gas',
                 moisture_ID=None):
        super().__init__(ID, ins, outs, thermo)
        self._isplit = self.chemicals.isplit(split)
        self.define_utility('Natural gas', self.natural_gas)
        self.P = P
        self.T = T
        self.R = R
        self.H = H
        self.length_to_diameter = length_to_diameter
        self.moisture_content = moisture_content
        self.utility_agent = utility_agent
        self.moisture_ID = moisture_ID
        
    @property
    def utility_agent(self):
        return self._utility_agent
    
    @utility_agent.setter
    def utility_agent(self, utility_agent):
        if utility_agent == 'Natural gas':
            pass
        elif utility_agent == 'Steam':
            self.heat_utilities = (bst.HeatUtility(),)
        else:
            raise ValueError(f"utility agent must be either 'Steam' or 'Natural gas'; not '{utility_agent}'")
        self._utility_agent = utility_agent
        
    def _run(self):
        wet_solids, air, natural_gas = self.ins
        dry_solids, hot_air, emissions = self.outs
        wet_solids.split_to(hot_air, dry_solids, self.split)
        sep.adjust_moisture_content(dry_solids, hot_air, self.moisture_content, self.moisture_ID)
        hot_air.P = air.P = self.P
        emissions.phase = air.phase = natural_gas.phase = hot_air.phase = 'g'
        design_results = self.design_results
        design_results['Evaporation'] = evaporation = hot_air.F_mass
        air.imass['N2', 'O2'] = np.array([0.78, 0.32]) * self.R * evaporation
        hot_air.mol += air.mol
        dry_solids.T = hot_air.T = self.T
        emissions.T = self.T + 30.
        natural_gas.empty()
        emissions.empty()
        if self.utility_agent == 'Natural gas':
            LHV = self.chemicals.CH4.LHV
            def f(CH4):
                CO2 = CH4    
                H2O = 2. * CH4
                natural_gas.imol['CH4'] = CH4
                emissions.imol['CO2', 'H2O'] = [CO2, H2O]    
                duty = (dry_solids.H + hot_air.H + emissions.H) - (wet_solids.H + air.H + natural_gas.H)
                CH4 = duty / LHV
                return CH4
            flx.wegstein(f, 0., 1e-3)
        
    def _design(self):
        length_to_diameter = self.length_to_diameter
        design_results = self.design_results
        design_results['Volume'] = volume = design_results['Evaporation'] / self.H 
        design_results['Diameter'] = diameter = cylinder_diameter_from_volume(volume, length_to_diameter)
        design_results['Length'] = length = diameter * length_to_diameter
        design_results['Peripheral drum area'] = cylinder_area(diameter, length)
        if self.utility_agent == 'Steam':
            self.heat_utilities[0](self.H_out - self.H_in, self.T)
        
        
class ThermalOxidizer(Unit):
    """
    Create a ThermalOxidizer that burns any remaining combustibles.
    
    Parameters
    ----------
    ins : stream sequence
        [0] Feed gas
        [1] Air
        [2] Natural gas
    outs : stream
        Emissions.
    tau : float, optional
        Residence time [hr]. Defaults to 0.00014 (0.5 seconds).
    duty_per_kg : float, optional
        Duty per kg of feed. Defaults to 105858 kJ / kg.
    V_wf : float, optional
        Fraction of working volume. Defaults to 0.95.
    
    Notes
    -----
    Adiabatic operation is assumed. Simulation and cost is based on [1]_.
    
    References
    ----------
    .. [1] Kwiatkowski, J. R.; McAloon, A. J.; Taylor, F.; Johnston, D. B. 
        Modeling the Process and Costs of Fuel Ethanol Production by the Corn 
        Dry-Grind Process. Industrial Crops and Products 2006, 23 (3), 288–296.
        https://doi.org/10.1016/j.indcrop.2005.08.004.

    """
    _N_ins = 3
    _N_outs = 1
    max_volume = 20. # m3
    _F_BM_default = {'Vessels': 2.06} # Assume same as dryer
    
    @property
    def natural_gas(self):
        """[Stream] Natural gas to satisfy steam and electricity requirements."""
        return self.ins[2]
    
    def __init__(self, *args, tau=0.00014, duty_per_kg=61., V_wf=0.95, 
                 **kwargs):
        Unit.__init__(self, *args, **kwargs)
        self.define_utility('Natural gas', self.natural_gas)
        self.tau = tau
        self.duty_per_kg = duty_per_kg
        self.V_wf = V_wf

    def _run(self):
        feed, air, ng = self.ins
        ng.imol['CH4'] = self.duty_per_kg * feed.F_mass / self.chemicals.CH4.LHV
        emissions, = self.outs
        emissions.mix_from([feed, ng])
        combustion_rxns = self.chemicals.get_combustion_reactions()
        combustion_rxns.force_reaction(emissions)
        O2 = max(-emissions.imol['O2'], 0.)
        air.imol['N2', 'O2'] = [0.78/0.32 * O2, O2]
        emissions.mix_from(self.ins)
        combustion_rxns.adiabatic_reaction(emissions)
        
    def _design(self):
        design_results = self.design_results
        volume = self.tau * self.outs[0].F_vol / self.V_wf
        V_max = self.max_volume
        design_results['Number of vessels'] = N = np.ceil(volume / V_max)
        design_results['Vessel volume'] = volume / N
        design_results['Total volume'] = volume
        
    def _cost(self):
        design_results = self.design_results
        total_volume = design_results['Total volume']
        N = design_results['Number of vessels']
        vessel_volume = design_results['Vessel volume']
        C = self.baseline_purchase_costs
        C['Vessels'] = N * 918300. * (vessel_volume / 13.18)**0.6


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import pytest
import biosteam as bst
from numpy.testing import assert_allclose
from thermo import SRK, PR


def test_isentropic_hydrogen_compressor():
    bst.settings.set_thermo(["H2"], cache=True)
    eta = 0.7
    feed = bst.Stream(H2=1, T=25 + 273.15, P=101325, phase='g')
    K = bst.units.IsentropicCompressor(ins=feed, P=50e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, 1151.3251608356125, 50e5],
    )
    # check compressor design
    assert K.design_results["Type"] == "Centrifugal"
    assert K.design_results["Driver"] == "Steam turbine"
    ideal_power = 4.919879203671207
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Steam turbine"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, 0, 1],
    )
    pass


def test_isentropic_two_phase_steam_compressor():
    bst.settings.set_thermo(["H2O"], cache=True)
    # feed is steam-water-mixture
    feed = bst.MultiStream(T=372.75, P=1e5, l=[('H2O', 0.1)], g=[('H2O', 0.9)])
    # from compress until superheated steam (test case for vle parameter)
    eta = 1
    K = bst.units.IsentropicCompressor(ins=feed, P=100e5, eta=eta, vle=True)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, 797.7528062886108, 100e5],
    )
    # check compressor design
    assert K.design_results["Type"] == "Centrifugal"
    assert K.design_results["Driver"] == "Steam turbine"
    ideal_power = 5.41038996517047
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Steam turbine"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, 0, 1],
    )
    pass


def test_isothermal_hydrogen_compressor():
    thermo = bst.Thermo([bst.Chemical('H2', eos=SRK)])
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream(H2=1, T=298.15, P=20e5, phase='g')

    ideal_power = 2.1024573963797715
    ideal_duty = -7261.402248757514

    # eta = 1
    eta = 1
    K = bst.units.IsothermalCompressor(ins=feed, P=350e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, feed.T, 350e5],
    )
    # check compressor design
    assert K.design_results["Type"] == "Reciprocating"
    assert K.design_results["Driver"] == "Electric motor"
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Electric motor"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, ideal_duty, 1],
    )
    # check heat utility
    expected_duty = ideal_duty / eta
    assert_allclose(
        [K.heat_utilities[0].unit_duty, K.heat_utilities[0].duty, K.heat_utilities[0].flow],
        [expected_duty, expected_duty, 7.525952491732307],
    )

    # repeat with eta=0.7
    eta = 0.7
    K = bst.units.IsothermalCompressor(ins=feed, P=350e5, eta=eta)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, feed.T, 350e5],
    )
    # check compressor design
    assert K.design_results["Type"] == "Reciprocating"
    assert K.design_results["Driver"] == "Electric motor"
    eta_motor = K.baseline_cost_algorithms[K.design_results["Type"]].efficiencies["Electric motor"]
    expected_power = ideal_power / eta / eta_motor
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Ideal power'],
           K.design_results['Ideal duty'], K.design_results['Compressors in parallel']],
        [expected_power, ideal_power, ideal_duty, 1],
    )
    # check heat utility
    expected_duty = ideal_duty / eta
    hu = K.heat_utilities[0]
    assert_allclose(
        [hu.unit_duty, hu.duty, hu.flow],
        [expected_duty, expected_duty, 7.525952491732307 / eta],
    )
    pass

def test_polytropic_hydrogen_compressor():
    thermo = bst.Thermo([bst.Chemical('H2', eos=PR)])
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream(H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    # eta=1 outlet state should be identical to ideal isentropic compression
    # schultz method
    eta = 1.0
    K = bst.units.PolytropicCompressor(ins=feed, P=P, eta=eta, method="schultz")
    K.simulate()
    out_poly = K.outs[0]
    K_isen = bst.units.IsentropicCompressor(ins=feed, P=P, eta=eta)
    K_isen.simulate()
    out_isen = K_isen.outs[0]
    assert K.heat_utilities == ()
    assert_allclose(
        [out_poly.T, out_poly.P, out_poly.H, K.power_utility.rate],
        [out_isen.T, out_isen.P, out_isen.H, K_isen.power_utility.rate],
    )
    # eta=1, hundseid method
    K = bst.units.PolytropicCompressor(ins=feed, P=P, eta=eta, method="hundseid")
    K.simulate()
    out_poly = K.outs[0]
    assert K.heat_utilities == ()
    assert_allclose(
        [out_poly.T, out_poly.P, out_poly.H, K.power_utility.rate],
        [out_isen.T, out_isen.P, out_isen.H, K_isen.power_utility.rate],
    )
    # eta=0.7, hundseid method
    eta=0.7
    feed = bst.Stream(H2=1, T=25 + 273.15, P=20e5, phase='g')
    K = bst.units.PolytropicCompressor(ins=feed, P=P, eta=eta, method="hundseid", n_steps=200)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, 958.1186, P],
    )
    # check compressor design
    assert K.design_results["Type"] == "Reciprocating"
    assert K.design_results["Driver"] == "Electric motor"
    assert K.heat_utilities == ()
    expected_power = 6.482427434951657
    actual_power = K.power_utility.rate
    assert_allclose(
        [actual_power, K.design_results['Compressors in parallel']],
        [expected_power, 1],
    )
    pass


def test_multistage_hydrogen_compressor_simple():
    thermo = bst.Thermo([bst.Chemical('H2', eos=PR)])
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream(H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    # test simple setup
    n_stages = 5
    pr = (P / feed.P) ** (1 / n_stages)
    K = bst.units.MultistageCompressor(ins=feed, pr=pr, n_stages=n_stages, eta=0.7)
    K.simulate()
    # check outlet state
    out = K.outs[0]
    assert_allclose(
        [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
        [1.0, 0.0, feed.T, P],
    )
    # check compressor design
    assert K.design_results["Type"] == "Multistage compressor"
    # assert K.design_results["Driver"] == "Electric motor"
    assert_allclose(
        [
            K.design_results["Area"],
            K.design_results['Tube side pressure drop'],
            K.design_results['Shell side pressure drop'],
        ],
        [1.7911402030788341, 15.0, 25.0],
    )
    # check heat utilities
    assert K.heat_utilities[0].ID == 'chilled_water'
    assert_allclose(
        [len(K.heat_utilities), K.heat_utilities[0].duty, K.heat_utilities[0].flow, K.heat_utilities[0].cost],
        [1, -11360.047353143065, 7.528681881161173, 0.056800236765715335],
    )
    # check power utility
    assert_allclose(
        [K.power_utility.consumption, K.power_utility.production, K.power_utility.rate, K.power_utility.cost],
        [4.668910747071128, 0.0, 4.668910747071128, 0.3651088204209622],
    )
    pass


def test_multistage_hydrogen_compressor_advanced():
    thermo = bst.Thermo([bst.Chemical('H2', eos=PR)])
    thermo.mixture.include_excess_energies = True
    bst.settings.set_thermo(thermo)
    feed = bst.Stream(H2=1, T=25 + 273.15, P=20e5, phase='g')
    P = 350e5
    # test advanced setup
    Ps = [60e5, 90e5, 200e5, 300e5, 350e5]
    Ts = [450, 400, 350, 300, feed.T]
    ks = [
        bst.units.IsentropicCompressor(P=Ps[0], eta=0.6),
        bst.units.PolytropicCompressor(P=Ps[1], eta=0.7),
        bst.units.IsentropicCompressor(P=Ps[2], eta=0.75),
        bst.units.PolytropicCompressor(P=Ps[3], eta=0.8),
        bst.units.IsentropicCompressor(P=Ps[4], eta=0.85),
    ]
    hxs = [bst.units.HXutility(T=T) for T in Ts]
    K = bst.units.MultistageCompressor(ins=feed, compressors=ks, hxs=hxs)
    K.simulate()
    # check hx outlet states
    for hx, P, T in zip(K.hxs, Ps, Ts):
        out = hx.outs[0]
        assert_allclose(
            [out.vapor_fraction, out.liquid_fraction, out.T, out.P],
            [1.0, 0.0, T, P],
        )
    # check compressor design
    assert K.design_results["Type"] == "Multistage compressor"
    # assert K.design_results["Driver"] == "Electric motor"
    assert_allclose(
        [
            K.design_results["Area"],
            K.design_results['Tube side pressure drop'],
            K.design_results['Shell side pressure drop'],
        ],
        [1.1011570059916886, 15.0, 25.0],
    )
    # check heat utilities
    assert len(K.heat_utilities) == 2
    assert K.heat_utilities[1].ID == 'chilled_water'
    assert K.heat_utilities[0].ID == 'cooling_water'
    assert_allclose(
        [K.heat_utilities[1].duty, K.heat_utilities[1].flow, K.heat_utilities[1].cost],
        [-3694.971298724524, 2.448780590727134, 0.018474856493622623],
    )
    assert_allclose(
        [K.heat_utilities[0].duty, K.heat_utilities[0].flow, K.heat_utilities[0].cost],
        [-10376.320213906532, 7.091122145383326, 0.0034594039386252554],
    )
    # check power utility
    assert_allclose(
        [K.power_utility.consumption, K.power_utility.production, K.power_utility.rate, K.power_utility.cost],
        [6.02039501950755, 0.0, 6.02039501950755, 0.47079489052549045],
    )
    pass

def test_compressor_design():
    bst.settings.set_thermo(["H2"], cache=True)
    feed = bst.Stream(H2=1, T=298.15, P=1e5, phase='g')
    K = bst.units.IsothermalCompressor(ins=feed, P=350e5)
    
    # Test default compressor types across pressures:
    psig_to_pascal = lambda psig: (psig + 14.6959) * 101325 / 14.6959
    def compressor_type_at_psig(psig):
        K.P = psig_to_pascal(psig)
        K.simulate()
        return K.design_results["Type"]
    
    # Pressures [psig] in the same order of magnitude as upper bound for each
    # compressor type.
    pressures = [1e2, 1e3, 1e4] 
    compressor_types = ['Screw', 'Centrifugal', 'Reciprocating']
    assert [compressor_type_at_psig(i) for i in pressures] == compressor_types
    
    # Test compressors in parallel:
    # Actual cubic feet per minute in the same order of magnitude as upper
    # bound for each compressory type.
    acfms = [1.5e4, 1e5, 5e3]
    
    def get_compressors_in_parallel(psig, acfm):
        K.ins[0].set_total_flow(acfm, 'cfm')
        K.P = psig_to_pascal(psig)
        K.simulate()
        return K.design_results['Compressors in parallel']
    
    compressors_in_parallel = [1 , 2]
    for psig, acfm in zip(pressures, acfms):
        assert compressors_in_parallel == [get_compressors_in_parallel(psig, i * acfm) for i in compressors_in_parallel]
        
    # Test user overwrite:
    K.P = psig_to_pascal(1e3) 
    K.compressor_type = "Centrifugal"
    K.driver = 'Gas turbine'
    K.driver_efficiency = 0.5
    K.simulate()
    assert K.design_results['Type'] == 'Centrifugal'
    assert K.design_results['Driver'] == 'Gas turbine'
    assert K.design_results['Driver efficiency'] == 0.5
    
    # Test capital cost correlations and factors
    def installed_equipment_cost_at_psig(psig):
        K.P = psig_to_pascal(psig)
        K.simulate()
        return K.installed_cost
    
    K.driver = K.driver_efficiency = K.compressor_type = 'Default'
    assert_allclose([installed_equipment_cost_at_psig(i) for i in pressures],
                    [ 2267264.256152,  5179043.514644, 10479970.002024])


if __name__ == '__main__':
    test_compressor_design()
    test_isentropic_hydrogen_compressor()
    test_isentropic_two_phase_steam_compressor()
    test_isothermal_hydrogen_compressor()
    test_polytropic_hydrogen_compressor()
    test_multistage_hydrogen_compressor_simple()
    test_multistage_hydrogen_compressor_advanced()



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from biorefineries.tests.test_biorefineries import (
    test_sugarcane,
    # test_lipidcane,
    test_cornstover,
    # test_LAOs
)

# To run these following tests, you'll need to get the github version.
# If you have the PyPI installation, these tests will not run.
# This is OK, the loss in coverage is very small.

# from biorefineries.tests import (
#     test_ethanol_adipic,
#     test_lactic
# )    

# Lactic acid tests failing for now due to some refractoring


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from __future__ import annotations
from thermosteam import settings
from thermosteam.utils import units_of_measure
from thermosteam.units_of_measure import (
    DisplayUnits, convert, power_utility_units_of_measure, AbsoluteUnitsOfMeasure
)
from typing import Optional

__all__ = ('PowerUtility',)


impact_indicator_basis = AbsoluteUnitsOfMeasure('kWhr')
default_price = 0.0782

@units_of_measure(power_utility_units_of_measure)
class PowerUtility:
    """
    Create an PowerUtility object that stores data on consumption and production
    of electricity.
            
    Notes
    -----
    The default price is 0.0782 USD/kWhr as suggested in [1]_.
    
    References
    ----------
    .. [1] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R.,
        & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
    
    Examples
    --------
    Create a PowerUtility object:
    
    >>> pu = PowerUtility()
    >>> pu
    PowerUtility(consumption=0.0, production=0.0)
    
    PowerUtility objects have `consumption` and `production` attributes
    which are updated when setting the rate with the assumption that
    a positive rate means no production (only consumption) and a negative
    rate means no consumption (only production).
    
    >>> pu(rate=-500)
    >>> pu.consumption, pu.production
    (0.0, 500.0)
    
    >>> pu(rate=500.)
    >>> pu.consumption, pu.production
    (500.0, 0.0)
    
    It is possible to have both consumption and production by setting these
    attributes individually (instead of setting rate)
    
    >>> pu.production = 100. 
    >>> pu.rate
    400.0
    
    Notice how the rate is equal to the consumption minus the production.
    
    The cost is available as a property:
    
    >>> pu.cost # USD/hr
    31.28
    
    It may be useful to print results in different units of measure:
    
    >>> pu.show(rate='BTU/s')
    PowerUtility:
     consumption: 474 BTU/s
     production: 94.8 BTU/s
     rate: 379 BTU/s
     cost: 31.3 USD/hr
    
    """
    __slots__ = ('consumption', 'production')
    
    #: Characterization factors for life cycle assessment [impact/kWhr] by impact key and kind (None, 'consumption', or 'production').
    characterization_factors: dict[tuple[str, str], float] = {}
    
    #: Units of measure for IPython display
    display_units: DisplayUnits = DisplayUnits(rate='kW', cost='USD/hr')
    
    def __init__(self, consumption: float=0., production: float=0.):
        #: Electricity consumption [kW]
        self.consumption: float = consumption
        
        #: Electricity production [kW]
        self.production: float = production
    
    def empty(self):
        """Set consumption and production to zero."""
        self.consumption = self.production = 0.
    
    @classmethod
    def get_CF(cls, key: str, consumption: Optional[bool]=True, 
               production: Optional[bool]=True, basis: Optional[str]=None, 
               units: Optional[str]=None):
        """
        Return the life-cycle characterization factor for consumption and 
        production on a kWhr basis given the impact key.
        
        Parameters
        ----------
        key :
            Name of impact indicator.
        consumption :
            Whether to return impact indicator for electricity consumption.
        production :
            Whether to return impact indicator for electricity production.
        basis :
            Basis of characterization factor. Energy is the only valid dimension. 
            Defaults to 'kWhr'.
        units :
            Units of impact indicator. Before using this argument, the default units 
            of the impact indicator should be defined with 
            :meth:`settings.define_impact_indicator <thermosteam._settings.ProcessSettings.define_impact_indicator>`.
            Units must also be dimensionally consistent with the default units.
        
        """
        try:
            value = cls.characterization_factors[key]
        except KeyError:
            value = (0., 0.)
        if consumption:
            if not production:
                value = value[0]
        elif production:
            value = value[1]
        else:
            return None
        if units is not None:
            original_units = settings.get_impact_indicator_units(key)
            f = original_units.conversion_factor(units)
            if consumption and production:
                value = (consumption * f, 
                         production * f)
            else:
                value *= f
        if basis is not None:
            f = impact_indicator_basis.conversion_factor(basis)
            consumption /= f
            production /= f
        return value

    @classmethod
    def set_CF(cls, key: str, consumption: Optional[float]=None,
               production: Optional[float]=None, basis: Optional[str]=None, 
               units: Optional[str]=None):
        """
        Set the life-cycle characterization factors for consumption and production
        on a kWhr basis given the impact key.
        
        Parameters
        ----------
        key :
            Name of impact indicator.
        consumption :
            Impact indicator for electricity consumption.
        production :
            Impact indicator for electricity production.
        basis :
            Basis of characterization factor. Energy is the only valid dimension. 
            Defaults to 'kWhr'.
        units :
            Units of impact indicator. Before using this argument, the default units 
            of the impact indicator should be defined with 
            :meth:`settings.define_impact_indicator <thermosteam._settings.ProcessSettings.define_impact_indicator>`.
            Units must also be dimensionally consistent with the default units.
        
        """
        if consumption is None:
            if production is None:
                raise ValueError("must pass at least one of either 'consumption' or 'production'")
            consumption = production
        elif production is None:
            production = consumption
        if units is not None:
            original_units = settings.get_impact_indicator_units(key)
            f = original_units.conversion_factor(units)
            consumption /= f 
            production /= f
        if basis is not None:
            f = impact_indicator_basis.conversion_factor(basis)
            consumption *= f
            production *= f
        cls.characterization_factors[key] = (consumption, production)
    
    @classmethod
    def default_price(cls):
        """Reset price back to BioSTEAM's default."""
        cls.price: float = default_price #: Electricity price [USD/kWhr]
    
    @property
    def rate(self) -> float:
        """Power requirement [kW]."""
        return self.consumption - self.production
    @rate.setter
    def rate(self, rate: float):
        rate = float(rate)
        if rate >= 0.:
            self.consumption = rate
            self.production = 0.
        else:
            self.consumption = 0.
            self.production = -rate
    
    @property
    def cost(self) -> float:
        """Cost [USD/hr]"""
        return self.price * self.rate
    
    def get_impact(self, key: str):
        """Return the impact [impact/hr] given characterization factor keys 
        for consumption and production. If no production key given, it defaults
        to the consumption key."""
        rate = self.consumption - self.production
        try: 
            cf = self.characterization_factors[key]
        except:
            return 0.
        return (cf[0] if rate > 0. else cf[1]) * rate
    
    def __bool__(self):
        return bool(self.consumption or self.production)
    
    def __call__(self, rate: float):
        """Set rate [kW]."""
        self.rate = rate
    
    def copy(self):
        return self.__class__(self.consumption, self.production)
    
    def mix_from(self, power_utilities: list[PowerUtility]):
        """
        Mix in requirements of power utilities.
        
        Examples
        --------
        >>> pus = [PowerUtility(production=100),
        ...        PowerUtility(consumption=50),
        ...        PowerUtility(production=20)]
        >>> pu = PowerUtility()
        >>> pu.mix_from(pus)
        >>> pu
        PowerUtility(consumption=50.0, production=120.0)
        
        """
        self.consumption = sum([i.consumption for i in power_utilities])
        self.production = sum([i.production for i in power_utilities])
    
    def copy_like(self, power_utility: PowerUtility):
        """Copy consumption anf production rates from another power utility."""
        self.consumption = power_utility.consumption
        self.production = power_utility.production
    
    def scale(self, scale: int):
        """Scale consumption and production accordingly."""
        self.consumption *= scale
        self.production *= scale
    
    @classmethod
    def sum(cls, power_utilities: list[PowerUtility]):
        """
        Return a PowerUtility object that represents the sum of power utilities.
        
        Examples
        --------
        >>> pus = [PowerUtility(production=100),
        ...        PowerUtility(consumption=50),
        ...        PowerUtility(production=20)]
        >>> pu = PowerUtility.sum(pus)
        >>> pu
        PowerUtility(consumption=50.0, production=120.0)
        
        """
        power_utility = cls()
        power_utility.mix_from(power_utilities)
        return power_utility
    
    def show(self, rate: Optional[str]=None, cost: Optional[str]=None):
        # Get units of measure
        display_units = self.display_units
        rate_units = rate or display_units.rate
        cost_units = cost or display_units.cost
        production = convert(self.production, 'kW', rate_units)
        consumption = convert(self.consumption, 'kW', rate_units)
        rate = consumption - production
        cost = convert(self.cost, 'USD/hr', cost_units)
        print(f'{type(self).__name__}:\n'
              f' consumption: {consumption:.3g} {rate_units}\n'
              f' production: {production:.3g} {rate_units}\n'
              f' rate: {rate:.3g} {rate_units}\n'
              f' cost: {cost:.3g} {cost_units}')
    _ipython_display_ = show    
    
    def __repr__(self) -> str:
        return f'{type(self).__name__}(consumption={self.consumption}, production={self.production})'

    def __add__(self, other: PowerUtility) -> PowerUtility:
        if other == 0: return self # Special case to get Python built-in sum to work
        return PowerUtility.sum([self, other])

    def __radd__(self, other: PowerUtility) -> PowerUtility:
        return self.__add__(other)
    
PowerUtility.default_price()
settings.__class__.set_electricity_CF = PowerUtility.set_CF
del units_of_measure, AbsoluteUnitsOfMeasure


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
@author: yoelcp and sarangbhagwat
"""
import biosteam as bst
from .splitting import Splitter
from biosteam.units.design_tools import PressureVessel
from math import sqrt, pi, ceil
import numpy as np
from numba import njit
import flexsolve as flx

__all__ = ('AdsorptionColumnTSA',)


@njit(cache=True)
def f_recovery_objective(
        superficial_velocity, target_recovery, F_mass_adsorbate, F_vol_feed, cycle_time, drying_time,
        adsorbent_capacity, rho_adsorbent, void_fraction, length_unused, regeneration_velocity,
        K
    ):
    diameter = 2 * sqrt(F_vol_feed / (superficial_velocity * pi))
    area = pi * diameter * diameter / 4
    total_length = (
         cycle_time * F_mass_adsorbate / (adsorbent_capacity * rho_adsorbent * area)
    ) + length_unused # length of equilibrium section plus unused bed (LES + LUB)
    length = total_length / 2 # Size of each column
    F_vol_regen = area * regeneration_velocity
    if K is not None:
        vessel_volume = length * 0.25 * diameter * diameter
        void_volume = void_fraction * vessel_volume
        N_washes = ceil(F_vol_regen * (cycle_time - drying_time) / void_volume)
        solvent = void_volume * 1e6 # m3 -> mL
        adsorbent = 1000 * vessel_volume * rho_adsorbent # g
        total_adsorbate = adsorbate = adsorbent * adsorbent_capacity # g
        stages = ceil(length / 0.1) # 0.1 m per stage
        solvent /= stages
        adsorbate_arr = adsorbate * np.ones(stages) / stages
        adsorbent_stage = adsorbent  / stages
        for i in range(stages * N_washes):
            # R * y + A * x = total
            # K = y / x
            # R * y  + A * y / K = total
            # y = total / (R + A / K)
            adsorbate_collected = 0
            for j in range(stages):
                y = (adsorbate_arr[j] + adsorbate_collected) / (solvent + adsorbent_stage / K)
                adsorbate_recovered = y * solvent - adsorbate_collected
                adsorbate_arr[j] -= adsorbate_recovered
                adsorbate_collected += adsorbate_recovered
        recovery = 1 - adsorbate_arr.sum() / total_adsorbate
    
    if target_recovery is None:
        error = 0
    else:
        error = target_recovery - recovery
    return (error, diameter, area, length, 
            F_vol_regen, vessel_volume, void_volume,
            stages, N_washes, recovery)


class AdsorptionColumnTSA(PressureVessel, Splitter):
    """
    Create a temperature swing adsorption (TSA) column. Default parameters
    are heuristic values for adsorption of water and polar components using
    silica gel. The design and cost algorithm of the adsorption column is 
    based on [1]_, [2]_, and [4]_. Three vessels are used: a lead, a guard, 
    and a standby vessel. The lead is in equilibrium, guard holds the 
    breakthrough curve and mass transfer zone, and the standby is in regeneration.
    Once the standby vessel is regenerated, it becomes the guard, the lead becomes
    the standby, and the guard becomes the lead. Therefore, the cycle time is
    the regeneration time.
    
    Parameters
    ----------
    ins : stream sequence
        * [0] Feed
        * [1] Regeneration fluid
    outs : stream sequence
        * [0] Effluent
        * [1] Purge
    superficial_velocity : float, optional
        Superficial velocity of the feed. The diameter of the receiving vessel adjusts
        accordingly. Defaults to 7.2 [m / hr]. Typical velocities are 4 to 14.4 m / hr for liquids [1]_.
    regeneration_velocity : float, optional
        Mean velocity of the fluid used to regenerate the bed. Defaults to 540 [m / hr]. 
        Common velocity range for gasses is 504-2160 m / hr [1]_.
    cycle_time : float, optional
        Time at which the receiving vessel is switched. Defaults to 3 [hr]. 
        Note that 1-2 hours required for thermal-swing-adsorption 
        (TSA) for silica gels. One hr is added to be conservative [2]_. 
    rho_adsorbent : float, optional
        The density of the adsorbent. Defaults to 480 [kg/m3], which is common 
        for silica gels [3]_.
    adsorbent_capacity : float, optional
        Fraction of absorbate that the absorbent can hold. Defaults to 0.1, a 
        conservative heuristic from [2]_.
    T_regeneration : float, optional
        Temperature during the regeneration phase. Defaults to 393.15 [K], which
        is used for silica gels [2]_.
    vessel_material : float, optional
        Vessel material. Defaults to 'Stainless steel 316',
    vessel_type : float, optional
        Vessel type. Defaults to 'Vertical'.
    adsorbate_ID : string
        Name of adsorbate.
    order : tuple[str], optional
        Order of component splits.
    regeneration_fluid : dict[str, float]
        Arguments to initialize fluid used to regenerate the bed.
    split : dict[str, float] or list[float], optional
        Component splits towards the effluent (0th outlet).
    
    Examples
    --------
    >>> import biosteam as bst
    >>> bst.default()
    >>> bst.settings.set_thermo(['Water', 'O2', 'N2', 'Hexane'], cache=True)
    >>> feed = bst.Stream('feed', Hexane=0.9, Water=0.1, units='kg/hr', total_flow=1000.)
    >>> A1 = bst.AdsorptionColumnTSA('A1', [feed, 'air'], 
    ...     split=dict(Water=0., Hexane=1.0),
    ...     adsorbate_ID='Water',
    ...  )
    >>> A1.simulate()
    >>> A1.results(with_units=False)
    Low pressure steam   Duty                                                    1.68e+03
                         Flow                                                      0.0433
                         Cost                                                      0.0103
    Design               Number of reactors                                             3
                         Vessel type                                             Vertical
                         Length                                                      52.1
                         Diameter                                                    1.67
                         Weight                                                   2.9e+03
                         Wall thickness                                              0.25
                         Vessel material                              Stainless steel 316
    Purchase cost        Heat exchanger regeneration - Double pipe                    117
                         Vertical pressure vessel                                1.47e+05
                         Platform and ladders                                    2.95e+04
                         Silica gel                                              7.23e+04
    Total purchase cost                                                          2.49e+05
    Utility cost                                                                   0.0103
    Name: A1, dtype: object
    
    References
    ----------
    [1] Adsorption basics Alan Gabelman (2017) Adsorption basics Part 1. AICHE
    [2] Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).
    [3] https://www.daisogelusa.com/technical-notes/approximate-packing-density-for-daisogel-bulk-silica-gel/
    [4] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani,
        R., & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
        Cost Accounting and Capital Cost Estimation (Chapter 16)
    
    """
    auxiliary_unit_names = ('heat_exchanger_regeneration',
                            'heat_exchanger_drying')
    
    # In $/ft3
    adsorbent_cost = {
        'Activated alumina': 72,
        'Activated carbon': 41,
        'Silica gel': 210,
        'Molecular sieves': 85,
    }
    
    # TODO: Update later plus ref
    _default_equipment_lifetime = {
        'Activated alumina': 10,
        'Activated carbon': 10,
        'Silica gel': 10,
        'Molecular sieves': 10,
    }
    
    # NOTE: Unit ignores cost of compressing 2.
    _N_ins = 3
    _N_outs = 3
    _N_heat_utilities = 1
    
    def __init__(self, 
            ID='', ins=None, outs=(), thermo=None, *,
            superficial_velocity=7.2, # m / hr; typical velocities are 4 to 14.4 m /hr for liquids; Adsorption basics Alan Gabelman (2017) Adsorption basics Part 1. AICHE
            regeneration_velocity=1332, # Mid point in velocity range for gasses, m / hr; Alan Gabelman (2017) Adsorption basics Part 1. AICHE
            cycle_time=3, # 1-2 hours required for thermal-swing-adsorption (TSA) for silica gels (add 1 hr for conservativeness); Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).
            rho_adsorbent=480, # Bulk density, including void fraction (in kg/m3) Common for silica gels https://www.daisogelusa.com/technical-notes/approximate-packing-density-for-daisogel-bulk-silica-gel/
            rho_adsorbent_solid = None, # Solid density, excluding void fraction (in kg/m3); used only if rho_adsorbent = None or False or 0
            adsorbent_capacity=0.1, # Conservative heuristic from Seider et. al. (2017) Product and Process Design Principles. Wiley
            T_regeneration=30 + 273.15, # For silica gels; Seader, J. D., Separation Process Principles: Chemical and Biochemical Operations,” 3rd ed., Wiley, Hoboken, NJ (2011).
            vessel_material='Stainless steel 316',
            vessel_type='Vertical',
            regeneration_fluid=dict(N2=0.78, O2=0.32, phase='g', units='kg/hr'),
            void_fraction=0.47, # Only matters when K given; 0.30 - 0.35 for activated carbon
            length_unused=1.219, # Additional length of a column to account for mass transfer limitations (due to unused bed). Defaults to +2 ft per column.
            drying_time=0., # Time for drying after regeneration
            T_air = 100 + 273.15,
            adsorbent='Silica gel',
            air_velocity = 1332,
            target_recovery=None,
            K=None,
            converge_adsorption_recovery=False,
            adsorbate_ID, 
            order=None, 
            wet_retention=0.01,
            split,
        ):
        bst.Splitter.__init__(self, ID, ins, outs, thermo, order=order, split=split)
        self.superficial_velocity = superficial_velocity
        self.cycle_time = cycle_time
        self.adsorbent_capacity = adsorbent_capacity
        self.adsorbate_ID = adsorbate_ID
        self.vessel_material = vessel_material
        self.vessel_type = vessel_type
        self.T_regeneration = T_regeneration
        self.regeneration_velocity = regeneration_velocity
        self.regeneration_fluid = regeneration_fluid
        self.void_fraction = void_fraction
        self.length_unused = length_unused
        self.T_air = T_air
        self.air_velocity = air_velocity
        self.drying_time = drying_time
        self.converge_adsorption_recovery = converge_adsorption_recovery
        self.wet_retention = wet_retention
        self.target_recovery = target_recovery
        self.adsorbent = adsorbent
        self.K = K
        self.rho_adsorbent_solid = rho_adsorbent_solid
        self.rho_adsorbent = rho_adsorbent if rho_adsorbent else rho_adsorbent_solid * (1-void_fraction)
        self.heat_exchanger_regeneration = bst.HXutility(None, None, None, thermo=thermo)
        self.heat_exchanger_drying = bst.HXutility(None, None, None, thermo=thermo)
        self.heat_utilities = (
            *self.heat_exchanger_regeneration.heat_utilities,
            *self.heat_exchanger_drying.heat_utilities,
        )
        
    @property
    def effluent(self):
        return self.outs[0]
        
    @property
    def regeneration_purge(self):
        return self.outs[1]
    
    def _run(self):
        feed, regen, dry_air = self.ins
        effluent, purge, air_purge = self.outs 
        regen.empty()
        dry_air.empty()
        for i in self.outs: i.empty()

        feed.split_to(effluent, purge, self.split)
        F_vol_feed = feed.F_vol
        superficial_velocity = self.superficial_velocity
        adsorbate_ID = self.adsorbate_ID
        F_mass_adsorbate = purge.imass[adsorbate_ID]
        target_recovery = self.target_recovery
        K = self.K # (g adsorbate / mL solvent)  /  (g adsorbate / g adsorbent)
        if K is not None:
            args = (
                target_recovery, F_mass_adsorbate, 
                F_vol_feed, self.cycle_time, self.drying_time,
                self.adsorbent_capacity, self.rho_adsorbent, self.void_fraction, 
                self.length_unused, self.regeneration_velocity,
                K
            )
            f = lambda *args: f_recovery_objective(*args)[0]
            if target_recovery is not None:
                if (y1:=f(14.4, *args)) <= 0.:
                    self.superficial_velocity = superficial_velocity = 14.4 # typical velocities are 4 to 14.4 m /hr for liquids; Adsorption basics Alan Gabelman (2017) Adsorption basics Part 1. AICHE
                elif (y0:=f(0.1, *args)) >= 0.:
                    self.superficial_velocity = superficial_velocity = 0.1 # set minimum velocity
                else:
                    self.superficial_velocity = superficial_velocity = flx.IQ_interpolation(
                        f, 0.1, 14.4, 
                        # x=self.superficial_velocity,
                        # x=7.25,
                        x=9.20,
                        # xtol=1e-6, ytol=1e-9, 
                        xtol=1e-5, ytol=1e-7, 
                        y0=y0, y1=y1,
                        args=args,
                    )
            (_, self.diameter, self.area, self.length,  
            self._F_vol_regen, self.vessel_volume, self.void_volume,
            self.equilibrium_stages, self.N_washes, self.recovery) = f_recovery_objective(superficial_velocity, *args)
            T_original = regen.T
            regen.reset_flow(**self.regeneration_fluid)
            purge.T = regen.T = self.T_regeneration
            regen.F_vol = self._F_vol_regen
            regen.T = T_original
            TAL = feed.imol[adsorbate_ID]
            split = self.isplit[adsorbate_ID]
            self.actual_split = actual_split = 1 - (self.recovery * (1 - split))
            effluent.imol[adsorbate_ID] = actual_split * TAL
            purge.mol += regen.mol
            purge.imol[adsorbate_ID] = feed.imol[adsorbate_ID] - effluent.imol[adsorbate_ID]
        else:
            self.diameter = diameter = 2 * sqrt(F_vol_feed / (superficial_velocity * pi))
            self.area = area = pi * diameter * diameter / 4
            total_length = (
                self.cycle_time * F_mass_adsorbate / (self.adsorbent_capacity * self.rho_adsorbent * area)
            ) + self.length_unused # length of equilibrium section plus unused bed (LES + LUB)
            self.length = length = total_length / 2 # Size of each column
            self.vessel_volume = length * area
            T_original = regen.T
            regen.reset_flow(**self.regeneration_fluid)
            purge.T = regen.T = self.T_regeneration
            regen.F_vol = area * self.regeneration_velocity
            regen.T = T_original
        
        if self.drying_time:
            air_purge.empty()
            air_purge.T = self.T_air
            air_purge.P = 10 * 101325
            air_purge.imass['N2'] = 1
            air_purge.phase = dry_air.phase = 'g'
            air_purge.F_vol = self.area * self.air_velocity * self.drying_time / self.cycle_time
            dry_air.copy_like(air_purge)
            retained_ethanol_vol = self.wet_retention * self.void_volume 
            retained_ethanol_mol = retained_ethanol_vol * (regen.mol / regen.F_vol)
            ethanol_mol = retained_ethanol_mol / self.cycle_time
            air_purge.mol += ethanol_mol
            H_out = air_purge.H
            regen._property_cache.clear()
            H_in0 = (regen.H / regen.F_vol) * retained_ethanol_vol # H_solvent 
            try:
                dry_air.H = H_out - H_in0
            except:
                breakpoint()
            purge.mol -= ethanol_mol
    
    def _design(self):
        feed, regen, dry_air = self.ins
        design_results = self.design_results
        diameter = self.diameter
        length = self.length
        design_results['Number of reactors'] = 3
        design_results.update(
            self._vessel_design(
                feed.P * 0.000145038, # Pa to psi
                diameter * 3.28084, # m to ft
                length * 3.28084, # m to ft
            )
        )
        hxr = self.heat_exchanger_regeneration
        hxr.ins.empty()
        hxr.outs.empty()
        hxr.ins[0] = regen.copy()
        hxr.outs[0] = regen.copy()
        hxr.T = self.T_regeneration
        hxr.simulate()
        if self.drying_time:
            hxd = self.heat_exchanger_drying
            hxd.ins.empty()
            hxd.outs.empty()
            hxd.ins[0] = fresh_air = dry_air.copy()
            fresh_air.T = 298.15
            hxd.outs[0] = dry_air.copy()
            hxd.T = dry_air.T
            hxd.simulate()
        else:
            self.heat_exchanger_drying._setup()
    
    def _cost(self):
        design_results = self.design_results
        baseline_purchase_costs = self.baseline_purchase_costs
        baseline_purchase_costs.update(self._vessel_purchase_cost(
            design_results['Weight'], design_results['Diameter'], design_results['Length']))
        N_reactors = design_results['Number of reactors']
        for i, j in baseline_purchase_costs.items():
            baseline_purchase_costs[i] *= N_reactors
        baseline_purchase_costs[self.adsorbent] = N_reactors * 35.3147 * self.vessel_volume * self.adsorbent_cost[self.adsorbent]



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst

__all__ = (
    'create_facilities', 
    'create_all_facilities',
    'create_coheat_and_power_system',
)

def create_all_facilities(
    feedstock,
    CIP_over_feedstock=None,
    plant_air_over_feedstock=None,
    fire_water_over_feedstock=None,
    recycle_process_water_streams=None,
    treated_water_streams=None,
    CT=True,
    CWP=True,
    CIP=True,
    FWT=True,
    ADP=True,
    WWT=True,
    CHP=True,
    HXN=True,
    PWC=True,
    CT_kwargs=None,
    CWP_kwargs=None,
    CIP_kwargs=None,
    FWT_kwargs=None,
    ADP_kwargs=None,
    WWT_kwargs=None,
    CHP_kwargs=None,
    HXN_kwargs=None,
    PWC_kwargs=None,
    area=None,
    blowdown_recycle=False,
):
    return create_facilities(
        feedstock, CIP_over_feedstock, plant_air_over_feedstock, fire_water_over_feedstock,
        recycle_process_water_streams, treated_water_streams,
        CT=CT,
        CWP=CWP,
        CIP=CIP,
        FWT=FWT,
        ADP=ADP,
        WWT=WWT,
        CHP=CHP,
        HXN=HXN,
        PWC=PWC,
        CT_kwargs=CT_kwargs,
        CWP_kwargs=CWP_kwargs,
        CIP_kwargs=CIP_kwargs,
        FWT_kwargs=FWT_kwargs,
        ADP_kwargs=ADP_kwargs,
        WWT_kwargs=WWT_kwargs,
        CHP_kwargs=CHP_kwargs,
        HXN_kwargs=HXN_kwargs,
        PWC_kwargs=PWC_kwargs,
        area=area,
        blowdown_recycle=blowdown_recycle,
    )

def create_facilities(
        feedstock=None,
        CIP_over_feedstock=None,
        plant_air_over_feedstock=None,
        fire_water_over_feedstock=None,
        recycle_process_water_streams=None,
        treated_water_streams=None,
        CT=None,
        CWP=None,
        CIP=None,
        FWT=None,
        ADP=None,
        WWT=None,
        CHP=None,
        HXN=None,
        PWC=None,
        CT_kwargs=None,
        CWP_kwargs=None,
        CIP_kwargs=None,
        FWT_kwargs=None,
        ADP_kwargs=None,
        WWT_kwargs=None,
        CHP_kwargs=None,
        HXN_kwargs=None,
        PWC_kwargs=None,
        area=None,
        blowdown_recycle=False,
    ):
    kwargs = (CHP_kwargs, WWT_kwargs, CT_kwargs, CWP_kwargs, CIP_kwargs,
              FWT_kwargs, ADP_kwargs, HXN_kwargs, PWC_kwargs)
    (CHP_kwargs, WWT_kwargs, CT_kwargs, CWP_kwargs, CIP_kwargs,
     FWT_kwargs, ADP_kwargs, HXN_kwargs, PWC_kwargs) = kwargs = [
         {} if i is None else i for i in kwargs
    ]
    if area:
        for i in kwargs[:2]: 
            if 'area' not in i: i['area'] = area
        for i in kwargs[2:]: 
            if 'ID' not in i: i['ID'] = area
    if CT: bst.facilities.CoolingTower(**CT_kwargs)
    if CWP: bst.facilities.ChilledWaterPackage(**CWP_kwargs)
    if WWT: WWT = bst.create_wastewater_treatment_system(mockup=True, autopopulate=True, **WWT_kwargs)
    if HXN: bst.HeatExchangerNetwork(**HXN_kwargs)
    if recycle_process_water_streams is None: 
        streams = bst.get_streams_from_context_level(0)
        recycle_process_water_streams = [i for i in streams if i.isproduct() and 'process_water' in i.ID]
    if feedstock is not None:
        if CIP:
            CIP = bst.Stream('CIP', Water=126, units='kg/hr')
            CIP_package = bst.CIPpackage(ins=CIP, **CIP_kwargs)
            CIP_package.CIP_over_feedstock = 0.00121 if CIP_over_feedstock is None else CIP_over_feedstock
            @CIP_package.add_specification(run=True)
            def adjust_CIP(): CIP.imass['Water'] = feedstock.F_mass * CIP_package.CIP_over_feedstock
        if ADP:
            plant_air = bst.Stream('plant_air', N2=83333, units='kg/hr')
            ADP = bst.AirDistributionPackage(ins=plant_air, **ADP_kwargs)
            ADP.plant_air_over_feedstock = 0.8 if plant_air_over_feedstock is None else plant_air_over_feedstock
            @ADP.add_specification(run=True)
            def adjust_plant_air(): plant_air.imass['N2'] = feedstock.F_mass * ADP.plant_air_over_feedstock
        if FWT:
            fire_water = bst.Stream('fire_water', Water=8343, units='kg/hr')
            FT = bst.FireWaterTank(ins=fire_water, **FWT_kwargs)
            FT.fire_water_over_feedstock = 0.08 if fire_water_over_feedstock is None else fire_water_over_feedstock
            @FT.add_specification(run=True)
            def adjust_fire_water(): fire_water.imass['Water'] = feedstock.F_mass * FT.fire_water_over_feedstock
    if CHP:
        create_coheat_and_power_system(mockup=True, autopopulate=True, **CHP_kwargs)
    if blowdown_recycle:
        units = bst.main_flowsheet.unit.get_context_level(0)
        blowdown_to_wastewater = bst.Stream('blowdown_to_wastewater')
        bst.BlowdownMixer(
            area or '',
            [i.blowdown_water for i in units if hasattr(i, 'blowdown_water')],
            blowdown_to_wastewater
        )
    if PWC:
        process_water_mixer = bst.Mixer(area or '', ins=recycle_process_water_streams or '')
        process_water = process_water_mixer.outs[0]
        if treated_water_streams is None:
            units = bst.main_flowsheet.unit.get_context_level(0)
            treated_water_streams = [i.treated_water for i in units if hasattr(i, 'treated_water')]
        treated_water_mixer = bst.Mixer(area or '', ins=treated_water_streams)
        treated_water = treated_water_mixer.outs[0]
        bst.ProcessWaterCenter(ins=[treated_water, '', process_water], **PWC_kwargs)

@bst.SystemFactory(
    ID='CHP_sys',
    ins=['makeup_water', 'natural_gas', 'FGD_lime', 'boiler_chemicals'],
    outs=['emissions', 'blowdown', 'ash'],
    fixed_ins_size=False,
)
def create_coheat_and_power_system(
        ins, outs, combustible_slurries=None, combustible_gases=None, 
        autopopulate=None, **kwargs
    ):
    makeup_water, natural_gas, lime, boiler_chemicals, = ins
    
    slurry_mixer = bst.Mixer('slurry_mixer', ins=combustible_slurries or [])
    gas_mixer = bst.Mixer('gas_mixer', ins=combustible_gases or [])
    BT = bst.BoilerTurbogenerator(
        ins=[slurry_mixer-0, gas_mixer-0, 
             makeup_water, natural_gas, lime, boiler_chemicals],
        outs=outs,
        **kwargs,
    )
    BT.autopopulate = False if autopopulate is None else autopopulate
    
    @BT.add_specification(run=True)
    def autopopulate_combustibles():
        if BT.autopopulate and not slurry_mixer.ins and not gas_mixer.ins:
            streams = bst.FreeProductStreams(BT.system.streams)
            slurry_mixer.ins.extend(streams.combustible_slurries)
            gas_mixer.ins.extend(streams.combustible_gases)
            slurry_mixer.run_until(BT)
            gas_mixer.run_until(BT)
            BT._system.update_configuration()



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains unit operations for wastewater treatment of a 
cellulosic ethanol biorefinery as in [1]_.

.. contents:: :local:

Data
----
.. autodata:: biosteam.units.wastewater.non_digestables
    
Unit operations
---------------
.. autoclass:: biosteam.units.wastewater.AnaerobicDigestion 
.. autoclass:: biosteam.units.wastewater.AerobicDigestion
.. autoclass:: biosteam.units.wastewater.ReverseOsmosis
.. autoclass:: biosteam.units.wastewater.WastewaterSystemCost

Utilities
---------
.. autofunction:: biosteam.units.wastewater.get_digestable_organic_chemicals

System factories
----------------

.. autofunction:: biosteam.units.wastewater.create_wastewater_treatment_units
.. autodata:: biosteam.units.wastewater.create_wastewater_treatment_system

References
----------
.. [1] Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A.,
    Dudgeon, D. (2011). Process Design and Economics for Biochemical 
    Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid 
    Pretreatment and Enzymatic Hydrolysis of Corn Stover
    (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269

"""
from .._unit import Unit
from .decorators import cost
from .splitting import Splitter
from .solids_separation import SolidsSeparator
from biosteam.utils import remove_undefined_chemicals, default_chemical_dict
import biosteam as bst
import thermosteam as tmo
from thermosteam import (
    utils,
    settings,
    separations,
    Reaction as Rxn,
    ParallelReaction as PRxn,
)

__all__ = (
    'AnaerobicDigestion', 
    'AerobicDigestion', 
    'ReverseOsmosis',
    'WastewaterSystemCost',
    'get_digestable_organic_chemicals',
    'create_wastewater_treatment_system',
)


# %% Functional utilities

#: list[str] IDs for non-digestible components in wastewater
non_digestables = ['WWTsludge', 'Cellulose', 'Xylan', 'CellulaseNutrients',
                   'Mannan', 'Lignin', 'Galactan', 'Glucan', 'Acetate',
                   'Biomass', 'Arabinan', 'Tar', 'CO', 'CO2', 'CH4']

def get_digestable_organic_chemicals(chemicals):
    """
    Return a list of digestible organic chemical IDs.

    Parameters
    ----------
    chemicals : :class:`~thermosteam.Chemicals`
        Digestible organic chemicals will be retrieve from this parameter.

    """
    non_digestable_chemicals = set([chemicals[i] for i in non_digestables if i in chemicals])
    digestables = [i for i in chemicals if i not in non_digestable_chemicals]
    return [i for i in digestables if i.locked_state != 'g' and 'C' in i.atoms]

def anaerobic_digestion_reactions(
        chemicals, MW_sludge,
        biogas_CH4_fraction=0.51, # mol-CH4 / mol-biogas
        organics_to_biogas=0.86, # g-biogas / g-reactant
        organics_to_biomass=0.05, # g-biomass / g-reactant
        thermo=None,
    ):
    # Defaults are based on P49 in Humbird et al., 91% of organic components is destroyed,	
    # of which 86% is converted to biogas and 5% is converted to sludge,	
    # and the biogas is assumed to be 51% CH4 and 49% CO2 on a dry molar basis	
    MW_CH4 = 16.04246
    MW_CO2 = 44.0095
    x_CH4 = biogas_CH4_fraction
    x_CO2 = (1. - x_CH4)
    MW_biogas = x_CH4 * MW_CH4 +  x_CO2 * MW_CO2 # g-biogas / mol-biogas
    conversion = organics_to_biogas + organics_to_biomass # g-reacted / g-reactant
    f_biogas = organics_to_biogas / (conversion * MW_biogas) # mol-biogas / g-reacted
    f_sludge = organics_to_biomass / (conversion * MW_sludge) # mol-biomass / g-reacted
    f_CH4 = x_CH4 * f_biogas # mol-CH4 / g-reacted
    f_CO2 = x_CO2 * f_biogas # mol-CO2 / g-reacted
    thermo = settings.get_default_thermo(thermo)
    parsable_name = thermo.chemicals.get_parsable_synonym
    isvalid = utils.is_valid_ID
    def anaerobic_rxn(chemical):
        reactant = chemical.ID
        if not isvalid(reactant): reactant = parsable_name(reactant)
        if reactant == 'H2SO4':
            return Rxn("H2SO4 -> H2S + 2O2", 'H2SO4', 1.)
        else:
            MW_inv = 1. / chemical.MW	
            return Rxn(f'{MW_inv}{reactant} -> {f_CH4}CH4 + {f_CO2}CO2 + {f_sludge}WWTsludge',	
                       reactant, 0.91, chemicals=thermo.chemicals)	    
    
    return PRxn([anaerobic_rxn(i) for i in chemicals])

def aerobic_digestion_reactions(chemicals, MW_sludge, X_combustion=0.74, X_growth=0.22, thermo=None):
    # Based on P49 in Humbird et al. Defaults assume 96% of remaining soluble 
    # organic matter is removed after aerobic digestion, of which 74% is 
    # converted to water and CO2 and 22% to cell mass
    isvalid = utils.is_valid_ID
    thermo = settings.get_default_thermo(thermo)
    parsable_name = thermo.chemicals.get_parsable_synonym
    def growth(chemical):
        f = MW_sludge / chemical.MW
        reactant = chemical.ID
        if not isvalid(reactant): reactant = parsable_name(reactant)
        return Rxn(f"{f}{reactant} -> WWTsludge", reactant, X_growth, chemicals=thermo.chemicals)
    return PRxn([i.get_combustion_reaction(conversion=X_combustion) + growth(i)
                 for i in chemicals])

# %% Unit operations

@cost('Flow rate', 'Wastewater system', units='kg/hr', CE=551,
      cost=50280080., n=0.6, BM=1, kW=7139/1.05, S=393100)
class WastewaterSystemCost(Unit): 
    """
    Create a unit that estimates the capital cost and electricity demand
    of a wastewater treatment system.
    
    Parameters
    ----------
    ins : stream
        Wastewater.
        
    """


class AnaerobicDigestion(Unit):
    """
    Create an anaerobic digestion unit operation. The model is based on 
    stoichiometric reactions and a specified fraction of water evaporated.
    
    Parameters
    ----------
    reactions : ReactionSet, optional
        Anaerobic digestion reactions. Default assumes 91% of organic components 
        is destroyed, of which 86% is converted to biogas and 5% is converted to 
        sludge.	The biogas is assumed to be 51% CH4 and 49% CO2 on a dry molar 
        basis.
    sludge_split : Array, optional
        Split between wastewater and sludge.
    ins : stream sequence
        * [0] Wastewater
    outs : stream sequence
        * [0] Biogas
        * [1] Wastewater
        * [2] Sludge
    
    """
    purchase_cost = installation_cost = 0
    _N_ins = 1
    _N_outs = 3
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 reactions=None, sludge_split=None):
        Unit.__init__(self, ID, ins, outs, thermo)
        chemicals = self.chemicals
        if not reactions:
            digestables = get_digestable_organic_chemicals(chemicals)
            reactions = anaerobic_digestion_reactions(digestables, chemicals.WWTsludge.MW, thermo=self.thermo)
        self.reactions = reactions
        if sludge_split is None:
            sludge_split = dict(
                Water=0.07087,
                Ethanol=0.0625,
                Furfural=0.06667,
                Glycerol=0.07377,
                LacticAcid=0.07084,
                SuccinicAcid=0.07377,
                HNO3=0.0678,
                Denaturant=0.07377,
                DAP=0.0678,
                AmmoniumAcetate=0.07084,
                AmmoniumSulfate=0.0678,
                H2SO4=0.0678,
                NaNO3=0.0678,
                Oil=0.07377,
                HMF=0.06667,
                NH3=0.07048,
                Glucose=0.06667,
                Xylose=0.07609,
                Sucrose=0.06915,
                Mannose=0.06915,
                Galactose=0.06915,
                Arabinose=0.06915,
                Extract=0.07084,
                Tar=0.7473,
                CaO=0.7473,
                Ash=0.7473,
                NaOH=0.0678,
                Lignin=0.744,
                SolubleLignin=0.07084,
                GlucoseOligomer=0.07143,
                GalactoseOligomer=0.07143,
                MannoseOligomer=0.07143,
                XyloseOligomer=0.07143,
                ArabinoseOligomer=0.07143,
                Z_mobilis=0.7438,
                T_reesei=0.7438,
                Cellulose=0.76,
                Protein=0.7391,
                Enzyme=0.7391,
                Xylan=0.75,
                Xylitol=0.07377,
                Cellobiose=0.06915,
                DenaturedEnzyme=0.7391,
                Arabinan=1,
                Mannan=1,
                Galactan=1,
                WWTsludge=0.7438,
                Cellulase=0.07084
            )
            remove_undefined_chemicals(sludge_split, chemicals)
            default_chemical_dict(sludge_split, chemicals, 0.07087, 0.07087, 0.744)
        self.sludge_split = chemicals.isplit(sludge_split)
        self._load_components()
    
    def _load_components(self):
        self.multi_stream = tmo.MultiStream(thermo=self.thermo)
    
    def _run(self):
        feed, = self.ins
        biogas, waste, sludge = self.outs
        biogas.phase = 'g'
        biogas.T = waste.T = sludge.T = 35+273.15
        sludge.copy_flow(feed)
        self.reactions(sludge)
        self.multi_stream.empty()
        self.multi_stream.copy_flow(sludge)
        self.multi_stream['g'].receive_vent(self.multi_stream['l'], energy_balance=False)
        biogas.mol[:] = self.multi_stream.imol['g']
        liquid_mol = self.multi_stream.imol['l']
        sludge.mol[:] = liquid_mol * self.sludge_split.data
        waste.mol[:] = liquid_mol - sludge.mol
        biogas.receive_vent(waste)
        

class AerobicDigestion(Unit):
    """
    Create an aerobic digestion unit operation. Model is based on 
    stoichiometric reactions and a specified fraction of water evaporated.
    
    Parameters
    ----------
    ins : stream sequence
        * [0] Wastewater    
        * [1] Air    
        * [2] Caustic    
    outs : stream sequence
        * [0] Vent    
        * [1] Treated wastewater
    reactions : ReactionSet, optional
        Aerobic digestion reactions. Defaults assume 96% of remaining soluble 
        organic matter is removed after aerobic digestion, of which 74% is 
        converted to water and CO2 and 22% to cell mass.
    evaporation : float, optional
        Fraction of water evaporated. Defaults to 0.0113.
        
    """    
    _N_ins = 3
    _N_outs = 2
    purchase_cost = installation_cost = 0
    
    def __init__(self, ID='', ins=None, outs=(), *, reactions=None, evaporation=0.0113):
        Unit.__init__(self, ID, ins, outs)
        if not reactions:
            chemicals = self.chemicals
            digestables = get_digestable_organic_chemicals(self.chemicals)
            reactions = aerobic_digestion_reactions(digestables, chemicals.WWTsludge.MW, thermo=self.thermo)        
        self.reactions = reactions
        self.evaporation = evaporation
    
    def _run(self):
        waste, air, caustic = self._ins
        vent, water = self.outs
        vent.phase = 'g'
        water.copy_like(waste)
        water.mol[:] += caustic.mol
        self.reactions.force_reaction(water)
        O2 = - water.imass['O2']
        N2 = 0.78 / 0.22 * O2
        air.empty()
        air.imass['O2', 'N2'] += [1.5 * O2, 1.5 * N2]
        water.imol['O2'] = 0.
        water.imass['N2'] = N2
        vent.copy_flow(water, ('CO2', 'O2', 'N2'))
        water_index = self.chemicals.index('7732-18-5')
        vent.mol[water_index] = water.mol[water_index] * self.evaporation
        water.mol[:] -= vent.mol
        

# TODO: Use moisture content
class SludgeCentrifuge(SolidsSeparator):
    """
    Create a centrifuge to separate sludge. The model is based on 
    component splits.
    
    Parameters
    ----------
    ins : stream
        Inlet fluid to be split.
    outs : stream sequence
        * [0] Liquid
        * [1] Sludge
    split : Defaults to Should be one of the following
        * [float] The fraction of net feed in the 0th outlet stream
        * [array_like] Componentwise split of feed to 0th outlet stream
        * [dict] ID-split pairs of feed to 0th outlet stream
    order=None : Iterable[str], defaults to biosteam.settings.chemicals.IDs
        Chemical order of split.
    
    """
    purchase_cost = installation_cost = 0
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, 
                 split=None, order=None, moisture_content=0.46):
        self._load_thermo(thermo)
        if split is None:
            chemicals = self.chemicals
            split = dict(
                Furfural=1,
                Glycerol=0.8889,
                LacticAcid=0.935,
                SuccinicAcid=0.8889,
                HNO3=0.9344,
                Denaturant=0.8889,
                DAP=0.9344,
                AmmoniumAcetate=0.935,
                AmmoniumSulfate=0.9344,
                H2SO4=0.935,
                NaNO3=0.9344,
                Oil=0.8889,
                HMF=1,
                NH3=0.9388,
                H2S=0.9394,
                SO2=0.9394,
                CO2=0.9333,
                NO2=0.9394,
                NO=0.9394,
                CO=0.9394,
                Glucose=1,
                Xylose=1,
                Sucrose=0.9286,
                Mannose=0.9286,
                Galactose=0.9286,
                Arabinose=0.9286,
                Extract=0.935,
                Tar=0.05155,
                CaO=0.05155,
                Ash=0.05155,
                NaOH=0.9344,
                Lignin=0.04943,
                SolubleLignin=0.935,
                GlucoseOligomer=0.9,
                GalactoseOligomer=0.9,
                MannoseOligomer=0.9,
                XyloseOligomer=0.9,
                ArabinoseOligomer=0.9,
                Z_mobilis=0.04991,
                T_reesei=0.04991,
                Cellulose=0.03846,
                Protein=0.05455,
                Enzyme=0.05455,
                Xylitol=0.8889,
                Cellobiose=0.9286,
                DenaturedEnzyme=0.05455,
                WWTsludge=0.04991,
                Cellulase=0.935
            )
            remove_undefined_chemicals(split, chemicals)
            default_chemical_dict(split, chemicals, 0.9394, 0.9286, 0.04991)
        SolidsSeparator.__init__(
            self, ID, ins, outs, thermo, split=split, order=order,
            moisture_content=moisture_content, 
        )
        
    def _run(self):
        ins = self.ins
        retentate, permeate = self.outs # Filtrate, solids
        separations.mix_and_split(ins, retentate, permeate, self.split)
        separations.adjust_moisture_content(permeate, retentate, self.moisture_content, None)
        
# TODO: Split values seem arbitrary in NREL 2011 model, perhaps work on a better model
class MembraneBioreactor(Splitter):
    """
    Create a membrane bioreactor to clarify sludge. The model is based on 
    component splits.
    
    Parameters 
    ----------
    ins : stream
        Inlet fluid to be split.
    outs : stream sequence
        * [0] Liquid
        * [1] Sludge
    split : Defaults to Should be one of the following
        * [float] The fraction of net feed in the 0th outlet stream
        * [array_like] Componentwise split of feed to 0th outlet stream
        * [dict] ID-split pairs of feed to 0th outlet stream
    order=None : Iterable[str], defaults to biosteam.settings.chemicals.IDs
        Chemical order of split.
    
    """
    purchase_cost = installation_cost = 0
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *, 
                 split=None, order=None):
        self._load_thermo(thermo)
        if split is None:
            chemicals = self.chemicals
            split = dict(
                Water=0.1454,
                Glycerol=0.125,
                LacticAcid=0.145,
                SuccinicAcid=0.125,
                HNO3=0.1454,
                Denaturant=0.125,
                DAP=0.1454,
                AmmoniumAcetate=0.145,
                AmmoniumSulfate=0.1454,
                H2SO4=0.1454,
                NaNO3=0.1454,
                Oil=0.125,
                N2=0.1351,
                NH3=0.1579,
                O2=0.15,
                CO2=0.1364,
                Xylose=0.25,
                Sucrose=0.125,
                Mannose=0.125,
                Galactose=0.125,
                Arabinose=0.125,
                Extract=0.145,
                NaOH=0.1454,
                SolubleLignin=0.145,
                GlucoseOligomer=0.1429,
                GalactoseOligomer=0.1429,
                MannoseOligomer=0.1429,
                XyloseOligomer=0.1429,
                ArabinoseOligomer=0.1429,
                Xylitol=0.125,
                Cellobiose=0.125,
                Cellulase=0.145
            )
            remove_undefined_chemicals(split, chemicals)
            default_chemical_dict(split, chemicals, 0.15, 0.125, 0.145)
        Splitter.__init__(self, ID, ins, outs, thermo, split=split, order=order)
        
class ReverseOsmosis(Unit):
    """
    Create a reverse osmosis unit operation for recovering water from brine.
    The model is based on a fraction of water recovered.
    
    Parameters
    ----------
    ins : stream
        Inlet fluid to be split.
    outs : stream sequence
        * [0] Filtered water
        * [1] Brine
    water_recovery : float, optional
        Water recovered to 0th stream. Defaults to 0.987
    
    """
    _N_ins = 1
    _N_outs = 2
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 water_recovery=0.987):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.water_recovery = water_recovery
    
    @property
    def treated_water(self):
        return self.outs[0]
    
    def _run(self):
        feed, = self.ins
        water, brine = self.outs
        water.copy_thermal_condition(feed)
        brine.copy_like(feed)
        water_index = self.chemicals.index('7732-18-5')
        water_flow = brine.mol[water_index]
        water_recovered = self.water_recovery * water_flow
        water.mol[water_index] = water_recovered
        brine.mol[water_index] = water_flow - water_recovered


def create_wastewater_treatment_units(ins, outs, 
        NaOH_price=0.07476, autopopulate=None
    ):
    """
    Create units for wastewater treatment, including anaerobic and aerobic 
    digestion reactors, a membrane bioreactor, a sludge centrifuge, 
    and a reverse osmosis unit.
    
    Parameters
    ----------
    ins : streams
        Wastewater streams (without solids). Defaults to all product streams
        at run time that are not sold and cannot generate energy through combustion
        (i.e. streams that have no sink, no price, and a LHV less that 1 kJ / g).
    outs : stream sequence
        * [0] methane
        * [1] sludge
        * [2] treated_water
        * [3] waste_brine 
    NaOH_price : float, optional
        Price of NaOH in USD/kg. The default is 0.07476.
    autopopulate : bool, optional
        Whether to automatically add waste water streams.
    """
    methane, sludge, treated_water, waste_brine = outs
    air = bst.Stream('air_lagoon', O2=51061, N2=168162, phase='g', units='kg/hr')
    caustic = bst.Stream('caustic', Water=2252, NaOH=2252,
                     units='kg/hr', price=NaOH_price)
    wastewater_mixer = bst.Mixer('M601', ins or [])
    wastewater_mixer.autopopulate = False if autopopulate is None else autopopulate
    
    @wastewater_mixer.add_specification(run=True)
    def autopopulate_waste_streams():
        if wastewater_mixer.autopopulate and not wastewater_mixer.ins: 
            sys = wastewater_mixer.system
            streams = bst.FreeProductStreams(sys.streams)
            wastewater_mixer.ins.extend(streams.noncombustible_slurries)
            for i in sys.facilities:
                if isinstance(i, bst.BlowdownMixer): wastewater_mixer.ins.append(i.outs[0])
            wastewater_mixer.system.update_configuration()
            
    WWTC = WastewaterSystemCost('WWTC', wastewater_mixer-0)
    anaerobic_digestion = AnaerobicDigestion('R601', WWTC-0, (methane, '', ''))
    recycled_sludge_mixer = bst.Mixer('M602', (anaerobic_digestion-1, ''))
    
    caustic_over_waste = caustic.imol['Water', 'NaOH'] / 2544301
    air_over_waste = air.imol['O2', 'N2'] / 2544301
    air.mol[:] = 0.
    waste = recycled_sludge_mixer-0
    def update_aerobic_input_streams():
        waste, air, caustic = aerobic_digestion.ins
        F_mass_waste = waste.F_mass
        caustic.imol['Water', 'NaOH'] = F_mass_waste * caustic_over_waste
        air.imol['O2', 'N2'] = F_mass_waste * air_over_waste
        aerobic_digestion._run()
    
    aerobic_digestion = AerobicDigestion('R602', (waste, air, caustic),
                                         outs=('evaporated_water', ''))
    aerobic_digestion.specification = update_aerobic_input_streams
    membrane_bioreactor = MembraneBioreactor('S601', aerobic_digestion-1)
    sludge_splitter = bst.Splitter('S602', membrane_bioreactor-1, split=0.96)
    fresh_sludge_mixer = bst.Mixer('M603', (anaerobic_digestion-2, sludge_splitter-1))
    sludge_centrifuge = SludgeCentrifuge('S603', fresh_sludge_mixer-0, outs=('', sludge))
    bst.Mixer('M604', [sludge_splitter-0, sludge_centrifuge-0], 0-recycled_sludge_mixer)
    reverse_osmosis = ReverseOsmosis('S604', membrane_bioreactor-0,
                                     outs=(treated_water, waste_brine))
    

def default_wastewater_thermo():
    from biorefineries.cornstover import chemicals
    return chemicals

create_wastewater_treatment_system = bst.SystemFactory(
    f=create_wastewater_treatment_units,
    ID='wastewater_treatment_sys',
    outs=[dict(ID='methane'),
          dict(ID='sludge'),
          dict(ID='treated_water'),
          dict(ID='waste_brine')],
    fixed_ins_size=False,
    fthermo=default_wastewater_thermo,
)
"""
Return a system for wastewater treatment, which includes anaerobic and aerobic 
digestion reactors, a membrane bioreactor, a sludge centrifuge, and a reverse 
osmosis unit.
 
Parameters
----------
ID : str, optional
    Defaults to 'wastewater_treatment_sys'.
ins : streams, optional
    Wastewater streams (without solids).
outs : stream sequence
    * [0] methane
    * [1] sludge
    * [2] treated_water
    * [3] waste_brine 
mockup : bool, optional
    Whether to create a mock system.
area : int, optional
    If given, IDs of all units will follow the area naming convention as
    explained in :func:`~biosteam.process_tools.utils.rename_unit`.
udct : bool, optional
    Whether to also return dictionary of units with their original IDs as keys.
NaOH_price : float, optional
    Price of NaOH in USD/kg. The default is 0.15.

Examples
--------
>>> from biosteam import Stream, create_wastewater_treatment_system
>>> feed = Stream(
...     ID='wastewater', 
...     Water=2.634e+04, 
...     Ethanol=0.07225, 
...     AceticAcid=24.67, 
...     Furfural=6.206, 
...     Glycerol=1.784, 
...     LacticAcid=17.7, 
...     SuccinicAcid=3.472, 
...     DAP=1.001, 
...     AmmoniumSulfate=17.63, 
...     HMF=2.366, 
...     Glucose=2.816, 
...     Xylose=6.953, 
...     Arabinose=12.78, 
...     Extract=65.98, 
...     Ash=83.52, 
...     Lignin=1.659, 
...     SolubleLignin=4.202, 
...     GlucoseOligomer=6.796, 
...     GalactoseOligomer=0.01718, 
...     MannoseOligomer=0.009008, 
...     XyloseOligomer=2.878, 
...     ArabinoseOligomer=0.3508, 
...     Z_mobilis=0.6668, 
...     Protein=2.569, 
...     Glucan=0.1555, 
...     Xylan=0.06121, 
...     Xylitol=4.88, 
...     Cellobiose=0.9419, 
...     Arabinan=0.02242, 
...     Mannan=0.06448, 
...     Galactan=0.01504, 
...     Cellulase=25.4, 
...     units='kmol/hr'
... )
>>> wwt_sys = create_wastewater_treatment_system()
>>> wwt_sys.simulate()
>>> wwt_sys.show()
System: wastewater_treatment_sys
ins...
[0] wastewater
    phase: 'l', T: 298.15 K, P: 101325 Pa
    composition: Water              0.946
                 Ethanol            6.64e-06
                 AceticAcid         0.00295
                 Furfural           0.00119
                 Glycerol           0.000328
                 LacticAcid         0.00318
                 SuccinicAcid       0.000818
                 DAP                0.000264
                 AmmoniumSulfate    0.00465
                 HMF                0.000595
                 Glucose            0.00101
                 Xylose             0.00208
                 Arabinose          0.00383
                 Extract            0.0237
                 Ash                0.000167
                 Lignin             0.000503
                 SolubleLignin      0.00128
                 GlucoseOligomer    0.00244
                 GalactoseOligomer  6.17e-06
                 MannoseOligomer    3.24e-06
                 XyloseOligomer     0.000862
                 ArabinoseOligomer  0.000105
                 Z_mobilis          3.28e-05
                 Protein            0.000117
                 Glucan             5.03e-05
                 Xylan              1.61e-05
                 Xylitol            0.00148
                 Cellobiose         0.000643
                 Arabinan           5.91e-06
                 Mannan             2.09e-05
                 Galactan           4.86e-06
                 Cellulase          0.00122
                 -----------------  5.01e+05 kg/hr
outs...
[0] methane
    phase: 'g', T: 307.76 K, P: 101325 Pa
    composition: Water         0.0318
                 Ethanol       2.46e-07
                 AceticAcid    1.9e-05
                 Furfural      2.96e-05
                 Glycerol      1.43e-10
                 LacticAcid    8.25e-09
                 SuccinicAcid  2.88e-11
                 CH4           0.266
                 CO2           0.702
                 ------------  2.13e+04 kg/hr
[1] sludge
    phase: 'l', T: 307.88 K, P: 101325 Pa
    composition: Water              0.46
                 Ethanol            5.28e-07
                 AceticAcid         0.000268
                 Glycerol           4.82e-05
                 LacticAcid         0.000263
                 SuccinicAcid       0.00012
                 DAP                0.000994
                 AmmoniumSulfate    0.0175
                 SO2                5.93e-06
                 Arabinose          0.000339
                 Extract            0.00196
                 Ash                0.0257
                 NaOH               0.015
                 Lignin             0.0755
                 SolubleLignin      0.000105
                 GlucoseOligomer    0.000313
                 GalactoseOligomer  7.91e-07
                 MannoseOligomer    4.15e-07
                 XyloseOligomer     0.00011
                 ArabinoseOligomer  1.35e-05
                 Z_mobilis          0.000407
                 Protein            0.00144
                 Glucan             0.000235
                 Xylan              0.000225
                 Xylitol            0.000218
                 Cellobiose         5.7e-05
                 Arabinan           0.000103
                 Mannan             0.000362
                 Galactan           8.45e-05
                 WWTsludge          0.399
                 Cellulase          0.0001
                 -----------------  2.57e+03 kg/hr
[2] treated_water
    phase: 'l', T: 307.79 K, P: 101325 Pa
    composition: Water  1
                 -----  4.29e+05 kg/hr
[3] waste_brine
    phase: 'l', T: 307.79 K, P: 101325 Pa
    composition: Water              0.5
                 Ethanol            1.34e-07
                 AceticAcid         6.06e-05
                 Furfural           2.43e-05
                 Glycerol           6.72e-06
                 LacticAcid         7.58e-05
                 SuccinicAcid       1.68e-05
                 DAP                0.0112
                 AmmoniumSulfate    0.198
                 HMF                1.23e-05
                 SO2                9.78e-05
                 Glucose            2.42e-05
                 Xylose             8.56e-05
                 Arabinose          7.87e-05
                 Extract            0.000565
                 Ash                0.00263
                 NaOH               0.22
                 Lignin             0.00509
                 SolubleLignin      3.04e-05
                 GlucoseOligomer    5.72e-05
                 GalactoseOligomer  1.45e-07
                 MannoseOligomer    7.59e-08
                 XyloseOligomer     2.02e-05
                 ArabinoseOligomer  2.46e-06
                 Z_mobilis          1.98e-07
                 Protein            7.28e-07
                 Glucan             0.0021
                 Xylan              0.000639
                 Xylitol            3.03e-05
                 Cellobiose         1.32e-05
                 Arabinan           0.000229
                 Mannan             0.000809
                 Galactan           0.000189
                 WWTsludge          0.0579
                 Cellulase          2.9e-05
                 -----------------  1.13e+04 kg/hr

"""